# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    numpy, 
    tf, 
    L1L2, 
    Dropout,
)
from mlgb.tf.functions import (
    ActivationLayer, 
    InitializerLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'AllFieldBinaryInteractionLayer',
    'AttentionalFieldBinaryInteractionLayer',
    'TwoAllFieldBinaryInteractionLayer',
    'GroupedAllFieldWiseBinaryInteractionLayer',
]


class BinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank not in (2, 3):
            raise MLGBError

        self.fbi_if_keepdim = True if input_shape.rank == 2 else False
        self.fbi_weight = self.add_weight(
            name='fbi_weight',
            shape=input_shape.as_list()[1:],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = x * self.fbi_weight
        square_of_sum = tf.square(tf.reduce_sum(x, axis=1, keepdims=self.fbi_if_keepdim))
        sum_of_square = tf.reduce_sum(tf.square(x), axis=1, keepdims=self.fbi_if_keepdim)
        x = 0.5 * tf.subtract(square_of_sum, sum_of_square)  # (batch_size, embed_dim) or (batch_size, 1)
        return x


class HigherOrderBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_hofm_order=3, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if not (fbi_hofm_order >= 3):
            raise MLGBError

        self.fbi_hofm_order = fbi_hofm_order

        self.fm_fn = BinaryInteractionLayer(
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError
        if self.fbi_hofm_order > input_shape[1]:
            raise MLGBError

        self.fields_width = input_shape[1]
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x_fm = self.fm_fn(x)
        x = x * self.fm_fn.fbi_weight
        x_hofm = self.get_anova_kernel(x, self.fbi_hofm_order)
        x = tf.concat([x_fm, x_hofm], axis=1)
        return x

    def get_anova_kernel(self, x, hofm_order):
        a = tf.ones_like(x)
        a = tf.gather(a, indices=tf.constant([0] * (self.fields_width + 1), dtype=tf.int32), axis=1)  # (b, f+1, e)

        ak_list = []
        for i in range(hofm_order):
            a_i_0 = tf.zeros_like(x)[:, :i+1, :]
            a_i = x[:, i:, :] * a[:, i:-1, :]
            a_i = tf.concat([a_i_0, a_i], axis=1)
            a = tf.math.cumsum(a_i, axis=1)  # (b, f+1, e)

            if i >= 2:
                ak_list.append(a[:, -1, :])  # (b, e)

        ak = tf.concat(ak_list, axis=1)  # (b, (order-2)*e)
        return ak


class FieldBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if fbi_mode not in ('FFM', 'FwFM', 'PNN:inner_product', 'PNN:outer_product'):
            raise MLGBError

        self.fbi_mode = fbi_mode
        self.fbi_unit = fbi_unit  # fbi_weight__length_of_latent_vector
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, self.fields_width, self.embed_dim = input_shape
        self.product_width = int(self.fields_width * (self.fields_width - 1) // 2)
        self.ij_ids = numpy.array(numpy.triu_indices(n=self.fields_width, k=1)).T
        # self.ij_ids = numpy.array(list(itertools.combinations(range(self.fields_width), 2)))

        self.weight_shape_map = {
            'FwFM': [self.product_width],
            'FFM': [self.fields_width, self.embed_dim, self.fbi_unit],
            'PNN:inner_product': [self.product_width, self.embed_dim, self.fbi_unit],
            'PNN:outer_product': [self.product_width, self.embed_dim, self.embed_dim, self.fbi_unit],
        }
        self.fbi_weight = self.add_weight(
            name='fbi_weight',
            shape=self.weight_shape_map[self.fbi_mode],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_i = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)

        if self.fbi_mode == 'FFM':
            w_i = tf.gather(params=self.fbi_weight, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=0)
            w_j = tf.gather(params=self.fbi_weight, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=0)
            w_ij = w_i * w_j
        else:
            w_ij = self.fbi_weight

        if self.fbi_mode == 'FwFM':
            x_ij = x_i * x_j
            x = tf.einsum('bfe,f->be', x_ij, w_ij)  # (batch_size, embed_dim)
        elif self.fbi_mode in ('FFM', 'PNN:inner_product'):
            x_ij = x_i * x_j
            x = tf.einsum('bfe,feu->bu', x_ij, w_ij)  # (batch_size, self.fbi_unit)
        elif self.fbi_mode == 'PNN:outer_product':
            x_i = tf.expand_dims(x_i, axis=3)
            x_j = tf.expand_dims(x_j, axis=2)
            x_ij = x_i @ x_j  # (bfe1,bf1e->bfee)
            x = tf.einsum('bfij,fiju->bu', x_ij, w_ij)  # (batch_size, self.fbi_unit)
        return x


class FieldProductBothInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='PNN:both', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if fbi_mode != 'PNN:both':
            raise MLGBError

        self.ip_fn = FieldBinaryInteractionLayer(
            fbi_mode='PNN:inner_product',
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            seed=seed,
        )
        self.op_fn = FieldBinaryInteractionLayer(
            fbi_mode='PNN:outer_product',
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x_ip = self.ip_fn(x)
        x_op = self.op_fn(x)
        x = tf.concat([x_ip, x_op], axis=1)  # (batch_size, fbi_unit * 2)
        return x


class BilinearInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='Bilinear:field_interaction', fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if fbi_mode not in ('Bilinear:field_all', 'Bilinear:field_each', 'Bilinear:field_interaction', 'FEFM', 'FvFM', 'FmFM'):
            raise MLGBError

        self.fbi_mode = fbi_mode
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, self.fields_width, self.embed_dim = input_shape
        self.product_width = int(self.fields_width * (self.fields_width - 1) // 2)
        self.ij_ids = numpy.array(numpy.triu_indices(n=self.fields_width, k=1)).T

        self.weight_num_map = {
            'Bilinear:field_all': 1,
            'Bilinear:field_each': self.fields_width,
            'Bilinear:field_interaction': self.product_width,
            'FEFM': self.product_width,
            'FvFM': self.product_width,
            'FmFM': self.product_width,
        }

        if self.fbi_mode == 'FvFM':
            self.fbi_weight_shape = [self.weight_num_map[self.fbi_mode], self.embed_dim]
        else:
            self.fbi_weight_shape = [self.weight_num_map[self.fbi_mode], self.embed_dim, self.embed_dim]

        self.fbi_weight = self.add_weight(
            name='fbi_weight',
            shape=self.fbi_weight_shape,
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_i = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)

        if self.fbi_mode == 'Bilinear:field_all':
            w = tf.concat([self.fbi_weight] * self.product_width, axis=0)
        elif self.fbi_mode == 'Bilinear:field_each':
            w = tf.gather(params=self.fbi_weight, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=0)
        elif self.fbi_mode in ('Bilinear:field_interaction', 'FmFM', 'FvFM'):
            w = self.fbi_weight
        elif self.fbi_mode == 'FEFM':
            w1 = self.fbi_weight
            w2 = tf.transpose(self.fbi_weight, perm=[0, 2, 1])
            w = (w1 + w2) * 0.5  # symmetric matrix
        else:
            raise MLGBError

        if self.fbi_mode == 'FvFM':
            x_i_mid = x_i * w  # (bfe,fe->bfe)
        else:
            x_i_mid = tf.einsum('bfe,fee->bfe', x_i, w)

        if self.fbi_mode == 'FEFM':
            x = tf.einsum('bme,bne->bmn', x_i_mid, x_j)  # bmm, (batch_size, product_width, product_width)
        else:
            x = x_i_mid * x_j  # (batch_size, product_width, embed_dim)

        x = tf.reduce_sum(x, axis=2, keepdims=False)  # (batch_size, product_width)
        return x


class AttentionalFieldBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_afm_activation='relu', fbi_afm_dropout=0.0, fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 seed=None):
        super().__init__()
        self.fbi_unit = fbi_unit
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=fbi_afm_activation,
            seed=seed,
        ).get()
        self.activation_fn = ActivationLayer(fbi_afm_activation)
        self.drop_fn = Dropout(fbi_afm_dropout)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, self.fields_width, self.embed_dim = input_shape
        self.ij_ids = numpy.array(numpy.triu_indices(n=self.fields_width, k=1)).T

        self.att_weight = self.add_weight(
            name='attention_weight',
            shape=[self.embed_dim, self.fbi_unit],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.att_bias = self.add_weight(
            name='attention_bias',
            shape=[self.fbi_unit],
            initializer='zeros',
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.proj_h = self.add_weight(
            name='projection_h',
            shape=[self.fbi_unit, 1],
            initializer=self.attention_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.proj_p = self.add_weight(
            name='projection_p',
            shape=[self.embed_dim, 1],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_i = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)
        x = x_i * x_j  # (batch_size, product_width, embed_dim)

        w = x @ self.att_weight + self.att_bias  # (batch_size, product_width, att_unit)
        w = self.activation_fn(w)
        w = w @ self.proj_h  # (batch_size, product_width, 1)
        w = tf.nn.softmax(w, axis=1)

        x = x * w  # (batch_size, product_width, embed_dim)
        x = self.drop_fn(x)
        x = x @ self.proj_p  # (batch_size, product_width, 1)
        x = tf.squeeze(x, axis=2)  # (batch_size, product_width)
        return x


class AllFieldBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0, seed=None):
        super().__init__()
        self.bi_mode_list = ('FM', 'FM3D',)
        self.ho_model_list = ('HOFM',)
        self.at_model_list = ('AFM',)
        self.pb_mode_list = ('PNN:both',)
        self.fb_mode_list = ('FFM', 'FwFM', 'PNN:inner_product', 'PNN:outer_product')
        self.bl_mode_list = ('Bilinear:field_all', 'Bilinear:field_each', 'Bilinear:field_interaction', 'FEFM', 'FvFM', 'FmFM')
        self.fbi_mode_list = self.bi_mode_list + self.ho_model_list + self.at_model_list + self.pb_mode_list + self.fb_mode_list + self.bl_mode_list
        if fbi_mode not in self.fbi_mode_list:
            raise MLGBError

        self.fbi_mode = fbi_mode

        if fbi_mode in self.pb_mode_list:
            self.fbi_fn = FieldProductBothInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.fb_mode_list:
            self.fbi_fn = FieldBinaryInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.bl_mode_list:
            self.fbi_fn = BilinearInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.at_model_list:
            self.fbi_fn = AttentionalFieldBinaryInteractionLayer(
                fbi_afm_activation=fbi_afm_activation,
                fbi_afm_dropout=fbi_afm_dropout,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.ho_model_list:
            self.fbi_fn = HigherOrderBinaryInteractionLayer(
                fbi_hofm_order=fbi_hofm_order,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        else:
            self.fbi_fn = BinaryInteractionLayer(
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank not in (2, 3):
            raise MLGBError
        if input_shape.rank == 2 and self.fbi_mode != 'FM':
            raise MLGBError
        if input_shape.rank != 2 and self.fbi_mode == 'FM':
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = self.fbi_fn(x)
        return x


class TwoBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]

        if input_0_shape.rank not in (2, 3):
            raise MLGBError
        if input_0_shape.rank != input_1_shape.rank:
            raise MLGBError
        if input_0_shape.rank == 3 and (input_0_shape[2] != input_1_shape[2]):  # embed_dim
            raise MLGBError

        self.fbi_if_keepdim = True if input_0_shape.rank == 2 else False

        self.fields_i_width, self.fields_j_width = input_0_shape[1], input_1_shape[1]
        self.ij_ids = numpy.array(numpy.meshgrid(range(self.fields_i_width), range(self.fields_j_width))).T.reshape(-1, 2)
        # self.ij_ids = numpy.array([[i, j] for i in range(self.fields_i_width) for j in range(self.fields_j_width)])

        self.fbi_i_weight = self.add_weight(
            name='fbi_i_weight',
            shape=input_shape[0].as_list()[1:],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.fbi_j_weight = self.add_weight(
            name='fbi_j_weight',
            shape=input_shape[1].as_list()[1:],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_i, x_j = inputs

        x_i = x_i * self.fbi_j_weight
        x_j = x_j * self.fbi_j_weight
        x_i = tf.gather(params=x_i, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x_j, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)

        x = x_i * x_j
        x = tf.reduce_sum(x, axis=1, keepdims=self.fbi_if_keepdim)  # (batch_size, embed_dim) or (batch_size, 1)
        return x


class TwoFieldBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if fbi_mode not in ('FFM', 'FwFM', 'PNN:inner_product', 'PNN:outer_product'):
            raise MLGBError

        self.fbi_mode = fbi_mode
        self.fbi_unit = fbi_unit  # fbi_weight__length_of_latent_vector
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]

        if input_0_shape.rank != 3:
            raise MLGBError
        if input_0_shape.rank != input_1_shape.rank:
            raise MLGBError
        if input_0_shape.rank == 3 and (input_0_shape[2] != input_1_shape[2]):  # embed_dim
            raise MLGBError

        self.embed_dim = input_0_shape[2]
        self.fields_i_width, self.fields_j_width = input_0_shape[1], input_1_shape[1]
        self.product_width = int(self.fields_i_width * self.fields_j_width)
        self.ij_ids = numpy.array(numpy.meshgrid(range(self.fields_i_width), range(self.fields_j_width))).T.reshape(-1, 2)

        self.weight_shape_map = {
            'FwFM': [self.product_width],
            'FFM': [self.fields_i_width + self.fields_j_width, self.embed_dim, self.fbi_unit],
            'PNN:inner_product': [self.product_width, self.embed_dim, self.fbi_unit],
            'PNN:outer_product': [self.product_width, self.embed_dim, self.embed_dim, self.fbi_unit],
        }
        self.fbi_weight = self.add_weight(
            name='fbi_weight',
            shape=self.weight_shape_map[self.fbi_mode],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_i, x_j = inputs
        x_i = tf.gather(params=x_i, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x_j, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)

        if self.fbi_mode == 'FFM':
            w_i = tf.gather(params=self.fbi_weight, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=0)
            w_j = tf.gather(params=self.fbi_weight, indices=tf.constant(self.ij_ids[:, 1] + self.fields_i_width, dtype=tf.int32), axis=0)  # shift
            w_ij = w_i * w_j
        else:
            w_ij = self.fbi_weight

        if self.fbi_mode == 'FwFM':
            x_ij = x_i * x_j
            x = tf.einsum('bfe,f->be', x_ij, w_ij)  # (batch_size, embed_dim)
        elif self.fbi_mode in ('FFM', 'PNN:inner_product'):
            x_ij = x_i * x_j
            x = tf.einsum('bfe,feu->bu', x_ij, w_ij)  # (batch_size, self.fbi_unit)
        elif self.fbi_mode == 'PNN:outer_product':
            x_i = tf.expand_dims(x_i, axis=3)
            x_j = tf.expand_dims(x_j, axis=2)
            x_ij = x_i @ x_j  # (bfe1,bf1e->bfee)
            x = tf.einsum('bfij,fiju->bu', x_ij, w_ij)  # (batch_size, self.fbi_unit)
        else:
            raise MLGBError
        return x


class TwoFieldProductBothInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='PNN:both', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if fbi_mode != 'PNN:both':
            raise MLGBError

        self.ip_fn = TwoFieldBinaryInteractionLayer(
            fbi_mode='PNN:inner_product',
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            seed=seed,
        )
        self.op_fn = TwoFieldBinaryInteractionLayer(
            fbi_mode='PNN:outer_product',
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]

        if input_0_shape.rank != 3:
            raise MLGBError
        if input_0_shape.rank != input_1_shape.rank:
            raise MLGBError
        if input_0_shape.rank == 3 and (input_0_shape[2] != input_1_shape[2]):  # embed_dim
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_i, x_j = inputs

        x_ip = self.ip_fn([x_i, x_j])
        x_op = self.op_fn([x_i, x_j])
        x = tf.concat([x_ip, x_op], axis=1)  # (batch_size, fbi_unit * 2)
        return x


class TwoBilinearInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='Bilinear:field_interaction', fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        if fbi_mode not in ('Bilinear:field_all', 'Bilinear:field_each', 'Bilinear:field_interaction', 'FEFM', 'FvFM', 'FmFM'):
            raise MLGBError

        self.fbi_mode = fbi_mode
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]

        if input_0_shape.rank != 3:
            raise MLGBError
        if input_0_shape.rank != input_1_shape.rank:
            raise MLGBError
        if input_0_shape.rank == 3 and (input_0_shape[2] != input_1_shape[2]):  # embed_dim
            raise MLGBError

        self.embed_dim = input_0_shape[2]
        self.fields_i_width, self.fields_j_width = input_0_shape[1], input_1_shape[1]
        self.product_width = int(self.fields_i_width * self.fields_j_width)
        self.ij_ids = numpy.array(numpy.meshgrid(range(self.fields_i_width), range(self.fields_j_width))).T.reshape(-1, 2)

        self.weight_num_map = {
            'Bilinear:field_all': 1,
            'Bilinear:field_each': self.fields_width,
            'Bilinear:field_interaction': self.product_width,
            'FEFM': self.product_width,
            'FvFM': self.product_width,
            'FmFM': self.product_width,
        }

        if self.fbi_mode == 'FvFM':
            self.fbi_weight_shape = [self.weight_num_map[self.fbi_mode], self.embed_dim]
        else:
            self.fbi_weight_shape = [self.weight_num_map[self.fbi_mode], self.embed_dim, self.embed_dim]

        self.fbi_weight = self.add_weight(
            name='fbi_weight',
            shape=self.fbi_weight_shape,
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_i, x_j = inputs
        x_i = tf.gather(params=x_i, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x_j, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)

        if self.fbi_mode == 'Bilinear:field_all':
            w = tf.concat([self.fbi_weight] * self.product_width, axis=0)
        elif self.fbi_mode == 'Bilinear:field_each':
            w = tf.gather(params=self.fbi_weight, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=0)
        elif self.fbi_mode in ('Bilinear:field_interaction', 'FmFM'):
            w = self.fbi_weight
        elif self.fbi_mode == 'FEFM':
            w1 = self.fbi_weight
            w2 = tf.transpose(self.fbi_weight, perm=[0, 2, 1])
            w = (w1 + w2) * 0.5  # symmetric matrix
        else:
            raise MLGBError

        if self.fbi_mode == 'FvFM':
            x_i_mid = x_i * w  # (bfe,fe->bfe)
        else:
            x_i_mid = tf.einsum('bfe,fee->bfe', x_i, w)

        if self.fbi_mode == 'FEFM':
            x = tf.einsum('bme,bne->bmn', x_i_mid, x_j)  # bmm, (batch_size, product_width, product_width)
        else:
            x = x_i_mid * x_j  # (batch_size, product_width, embed_dim)

        x = tf.reduce_sum(x, axis=2, keepdims=False)  # (batch_size, product_width)
        return x


class TwoAttentionalFieldBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_afm_activation='relu', fbi_afm_dropout=0.0, fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 seed=None):
        super().__init__()
        self.fbi_unit = fbi_unit
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=fbi_afm_activation,
            seed=seed,
        ).get()
        self.activation_fn = ActivationLayer(fbi_afm_activation)
        self.dropout_fn = Dropout(fbi_afm_dropout)

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]

        if input_0_shape.rank != 3:
            raise MLGBError
        if input_0_shape.rank != input_1_shape.rank:
            raise MLGBError
        if input_0_shape.rank == 3 and (input_0_shape[2] != input_1_shape[2]):  # embed_dim
            raise MLGBError

        self.embed_dim = input_0_shape[2]
        self.fields_i_width, self.fields_j_width = input_0_shape[1], input_1_shape[1]
        self.product_width = int(self.fields_i_width * self.fields_j_width)
        self.ij_ids = numpy.array(numpy.meshgrid(range(self.fields_i_width), range(self.fields_j_width))).T.reshape(-1, 2)

        self.att_weight = self.add_weight(
            name='attention_weight',
            shape=[self.embed_dim, self.fbi_unit],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.att_bias = self.add_weight(
            name='attention_bias',
            shape=[self.fbi_unit],
            initializer='zeros',
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.proj_h = self.add_weight(
            name='projection_h',
            shape=[self.fbi_unit, 1],
            initializer=self.attention_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.proj_p = self.add_weight(
            name='projection_p',
            shape=[self.embed_dim, 1],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_i = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        x_j = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)
        x = x_i * x_j  # (batch_size, product_width, embed_dim)

        w = x @ self.att_weight + self.att_bias  # (batch_size, product_width, att_unit)
        w = self.activation_fn(w)
        w = w @ self.proj_h  # (batch_size, product_width, 1)
        w = tf.nn.softmax(w, axis=1)

        x = x * w  # (batch_size, product_width, embed_dim)
        x = self.drop_fn(x)
        x = x @ self.proj_p  # (batch_size, product_width, 1)
        x = tf.squeeze(x, axis=2)  # (batch_size, product_width)
        return x


class TwoAllFieldBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_afm_activation='relu', fbi_afm_dropout=0.0, seed=None):
        super().__init__()
        self.bi_mode_list = ('FM', 'FM3D',)
        self.at_mode_list = ('AFM',)
        self.pb_mode_list = ('PNN:both',)
        self.fb_mode_list = ('FFM', 'FwFM', 'PNN:inner_product', 'PNN:outer_product')
        self.bl_mode_list = ('Bilinear:field_all', 'Bilinear:field_each', 'Bilinear:field_interaction', 'FEFM', 'FvFM', 'FmFM')
        self.fbi_mode_list = self.bi_mode_list + self.at_model_list + self.pb_mode_list + self.fb_mode_list + self.bl_mode_list
        if fbi_mode not in self.fbi_mode_list:
            raise MLGBError

        self.fbi_mode = fbi_mode
        if fbi_mode in self.pb_mode_list:
            self.two_fbi_fn = TwoFieldProductBothInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.fb_mode_list:
            self.two_fbi_fn = TwoFieldBinaryInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.bl_mode_list:
            self.two_fbi_fn = TwoBilinearInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        elif fbi_mode in self.at_model_list:
            self.two_fbi_fn = TwoAttentionalFieldBinaryInteractionLayer(
                fbi_afm_activation=fbi_afm_activation,
                fbi_afm_dropout=fbi_afm_dropout,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )
        else:
            self.two_fbi_fn = TwoBinaryInteractionLayer(
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                seed=seed,
            )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]

        if input_0_shape.rank not in (2, 3):
            raise MLGBError
        if input_0_shape.rank != input_1_shape.rank:
            raise MLGBError
        if input_0_shape.rank == 3 and (input_0_shape[2] != input_1_shape[2]):  # embed_dim
            raise MLGBError
        if input_0_shape.rank == 2 and self.fbi_mode != 'FM':
            raise MLGBError
        if input_0_shape.rank != 2 and self.fbi_mode == 'FM':
            raise MLGBError

    @tf.function
    def call(self, inputs):
        x_i, x_j = inputs
        x = self.two_fbi_fn([x_i, x_j])
        return x


class GroupedAllFieldWiseBinaryInteractionLayer(tf.keras.layers.Layer):
    def __init__(self, group_indices=((0, 1), (2, 3), (4, 5, 6)),
                 fbi_fm_mode='FM3D', fbi_mf_mode='FwFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 seed=None):
        super().__init__()
        if group_indices and len(group_indices) < 2:
            raise MLGBError

        self.group_indices = group_indices  # (user, item, context)
        self.group_width = len(self.group_indices)
        self.product_width = int(self.group_width * (self.group_width - 1) // 2)
        self.ij_ids = numpy.array(numpy.triu_indices(n=self.group_width, k=1)).T

        self.fbi_fm_fn_list = [
            AllFieldBinaryInteractionLayer(
                fbi_mode=fbi_fm_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                fbi_afm_activation=fbi_afm_activation,
                fbi_afm_dropout=fbi_afm_dropout,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(self.group_width)
        ]
        self.fbi_mf_fn_list = [
            TwoAllFieldBinaryInteractionLayer(
                fbi_mode=fbi_mf_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                fbi_afm_activation=fbi_afm_activation,
                fbi_afm_dropout=fbi_afm_dropout,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(self.product_width)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        fbi_fm_pool, fbi_mf_pool = [], []
        for k in range(self.group_width):
            x_k = tf.gather(x, indices=tf.constant(self.group_indices[k], dtype=tf.int32), axis=1)
            fbi_fm_k_outputs = self.fbi_fm_fn_list[k](x_k)
            fbi_fm_pool.append(fbi_fm_k_outputs)

        for k, (i, j) in enumerate(self.ij_ids):
            x_i = tf.gather(x, indices=tf.constant(self.group_indices[i], dtype=tf.int32), axis=1)
            x_j = tf.gather(x, indices=tf.constant(self.group_indices[j], dtype=tf.int32), axis=1)
            fbi_mf_k_ouputs = self.fbi_mf_fn_list[k]([x_i, x_j])
            fbi_mf_pool.append(fbi_mf_k_ouputs)

        fbi_pool = fbi_fm_pool + fbi_mf_pool
        x = tf.concat(fbi_pool, axis=1)
        return x


























# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    numpy,
    tf,
    Dense,
    L1L2,
    Dropout,
    BatchNormalization,
    LayerNormalization,
    Flatten,
    Conv1D,
    Conv2D,
    MaxPool1D,
    MaxPool2D,
    GRU,
    BiGRUModeList,
)
from mlgb.tf.functions import (
    IdentityLayer,
    ActivationLayer,
    InitializerLayer,
    MaskLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'TaskLayer',
    'LinearLayer',
    'DeepNeuralNetworkLayer',
    'FeedForwardNetworkLayer',
    'Linear3dParallelLayer',
    'DNN3dParallelLayer',
    'Linear2dParallelLayer',
    'ConvolutionalNeuralNetworkLayer',
    'GatedRecurrentUnitLayer',
    'BiGatedRecurrentUnitLayer',
    'CapsuleNetworkLayer',
]


class TaskLayer(tf.keras.layers.Layer):
    def __init__(self, task='multiclass:10', task_multiclass_if_project=False, task_multiclass_if_softmax=False,
                 task_multiclass_temperature_ratio=None,
                 task_linear_if_identity=False, task_linear_if_weighted=False, task_linear_if_bias=True,
                 task_linear_initializer=None, seed=None):
        super().__init__()
        self.task = task
        self.task_name = task.split(':')[0]
        if self.task_name not in ('binary', 'regression', 'multiclass'):
            raise MLGBError
        if task_multiclass_temperature_ratio and not (1e-2 <= task_multiclass_temperature_ratio <= 1.0):
            raise MLGBError

        self.task_linear_if_identity = task_linear_if_identity
        self.task_linear_if_weighted = task_linear_if_weighted
        self.task_multiclass_if_project = task_multiclass_if_project
        self.task_multiclass_if_softmax = task_multiclass_if_softmax
        self.task_multiclass_temperature_ratio = task_multiclass_temperature_ratio
        self.task_activation_dict = {
            'regression': None,
            'binary': 'sigmoid',
            'multiclass': 'softmax',
        }
        self.activation_fn = ActivationLayer(self.task_activation_dict[self.task_name])
        self.task_linear_initializer = InitializerLayer(
            initializer=task_linear_initializer,
            activation=self.task_activation_dict[self.task_name],
            seed=seed,
        ).get()
        self.linear_fn = Dense(
            units=1 if self.task_name != 'multiclass' else int(task.split(':')[1]),
            activation=None,
            use_bias=task_linear_if_bias,
            kernel_initializer=self.task_linear_initializer,
        )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError
        if self.task_name == 'multiclass' and not self.task_multiclass_if_project:
            if int(self.task.split(':')[1]) != input_shape[1]:
                raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        if self.task_name != 'multiclass':
            if not self.task_linear_if_identity:
                if self.task_linear_if_weighted:
                    x = self.linear_fn(x)
                else:
                    x = tf.reduce_sum(x, axis=1, keepdims=True)  # only apply after product of matching model.
            x = self.activation_fn(x)
        else:
            if self.task_multiclass_if_project:
                x = self.linear_fn(x)

            if self.task_multiclass_temperature_ratio:
                x /= self.task_multiclass_temperature_ratio  # zoom data for better performance in softmax_loss.

            if self.task_multiclass_if_softmax:
                x = tf.nn.softmax(x, axis=1)
        return x


class LinearV1Layer(tf.keras.layers.Layer):
    def __init__(self, linear_l1=0.0, linear_l2=0.0, linear_initializer=None, linear_activation=None,
                 linear_if_bias=True, linear_unit=1, seed=None):
        super().__init__()
        self.linear_if_bias = linear_if_bias
        self.linear_l1 = linear_l1
        self.linear_l2 = linear_l2
        self.linear_unit = linear_unit  # linear_weight__length_of_latent_vector

        self.linear_initializer = InitializerLayer(
            initializer=linear_initializer,
            activation=linear_activation,
            seed=seed,
        ).get()
        self.activation_fn = ActivationLayer(linear_activation)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.linear_weight = self.add_weight(
            name='linear_weight',
            shape=[input_shape[-1], self.linear_unit],
            initializer=self.linear_initializer,
            regularizer=L1L2(self.linear_l1, self.linear_l2),
            trainable=True,
        )
        self.linear_bias = self.add_weight(
            name='linear_bias',
            shape=[self.linear_unit],
            initializer='zeros',
            regularizer=L1L2(self.linear_l1, self.linear_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = tf.matmul(x, self.linear_weight) + (self.linear_bias if self.linear_if_bias else 0.0)
        x = self.activation_fn(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'linear_weight': self.linear_weight.numpy(),
                'linear_bias': self.linear_bias.numpy(),
            }
        )
        return config


class LinearV2Layer(tf.keras.layers.Layer):
    # tf: default linear_initializer=None='glorot_uniform'; torch: default linear_initializer=None='he_uniform'.
    def __init__(self, linear_l1=0.0, linear_l2=0.0, linear_initializer=None, linear_activation=None,
                 linear_if_bias=True, linear_unit=1, seed=None):
        super().__init__()
        self.linear_unit = linear_unit
        self.linear_if_bias = linear_if_bias
        self.linear_activation = linear_activation
        self.linear_l1 = linear_l1
        self.linear_l2 = linear_l2
        self.linear_initializer = InitializerLayer(
            initializer=linear_initializer,
            activation=linear_activation,
            seed=seed,
        ).get()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        f = input_shape[1]
        self.linear_fn = Dense(
            units=self.linear_unit if self.linear_unit else f,
            activation=self.linear_activation,
            use_bias=self.linear_if_bias,
            kernel_initializer=self.linear_initializer,
            bias_initializer='zeros',
            kernel_regularizer=L1L2(self.linear_l1, self.linear_l2),
            bias_regularizer=L1L2(self.linear_l1, self.linear_l2),
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = self.linear_fn(x)
        return x


class LinearLayer(tf.keras.layers.Layer):
    def __init__(self, linear_activation=None, linear_if_bias=True, linear_l1=0.0, linear_l2=0.0,
                 linear_initializer=None, linear_unit=1, seed=None):
        super().__init__()
        self.linear_fn = LinearV2Layer(
            linear_unit=linear_unit,
            linear_activation=linear_activation,
            linear_if_bias=linear_if_bias,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        # tf.print('input_shape:', input_shape)
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = self.linear_fn(x)
        return x


class DeepNeuralNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None, dnn_if_bias=True, seed=None):
        super().__init__()
        self.dnn_hidden_units = dnn_hidden_units
        self.dnn_length = len(self.dnn_hidden_units)
        self.dnn_dropouts = [dnn_dropout] * self.dnn_length if isinstance(dnn_dropout, float) else dnn_dropout
        self.dnn_bns = [dnn_if_bn] * self.dnn_length if isinstance(dnn_if_bn, bool) else dnn_if_bn
        self.dnn_lns = [dnn_if_ln] * self.dnn_length if isinstance(dnn_if_ln, bool) else dnn_if_ln
        self.dnn_bias = [dnn_if_bias] * self.dnn_length if isinstance(dnn_if_bias, bool) else dnn_if_bias
        self.dnn_initializer_list = [
            InitializerLayer(
                initializer=dnn_initializer,
                activation=dnn_activation,
                seed=seed + i if isinstance(seed, int) else seed,
            ).get() for i in range(self.dnn_length)
        ]
        self.dnn_fn_list = [
            [
                Dense(
                    units=self.dnn_hidden_units[i],
                    kernel_initializer=self.dnn_initializer_list[i],
                    kernel_regularizer=L1L2(l1=0.0, l2=dnn_l2),
                    use_bias=self.dnn_bias[i],
                ),
                BatchNormalization(axis=1) if self.dnn_bns[i] else IdentityLayer(),
                LayerNormalization(axis=1) if self.dnn_lns[i] else IdentityLayer(),
                Dropout(rate=self.dnn_dropouts[i], seed=seed + i if isinstance(seed, int) else seed),
                ActivationLayer(activation=dnn_activation),
            ]
            for i in range(self.dnn_length)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        for dnn_fn in self.dnn_fn_list:
            for fn in dnn_fn:
                x = fn(x)
        return x


class FeedForwardNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, ffn_linear_if_twice=True, ffn_if_bias=True, ffn_activation='relu',
                 ffn_dropout=0.0, ffn_if_bn=False, ffn_if_ln=False, ffn_l2=0.0, ffn_initializer=None,
                 ffn_last_factor=None, ffn_last_activation=None,
                 seed=None):
        super().__init__()
        self.ffn_linear_if_twice = ffn_linear_if_twice
        self.ffn_if_bias = ffn_if_bias
        self.ffn_l2 = ffn_l2
        self.ffn_last_factor = ffn_last_factor
        self.ffn_last_activation = ffn_last_activation

        self.bn_fn = BatchNormalization(axis=1) if ffn_if_bn else IdentityLayer()
        self.ln_fn = LayerNormalization(axis=1) if ffn_if_ln else IdentityLayer()
        self.drop_fn = Dropout(rate=ffn_dropout, seed=seed)
        self.ffn_initializer_list = [
            InitializerLayer(
                initializer=ffn_initializer,
                activation=None,
                seed=seed + i if isinstance(seed, int) else seed,
            ).get()
            for i in range(2 if self.ffn_linear_if_twice else 1)
        ]
        self.activation_fn = ActivationLayer(activation=ffn_activation)
        if self.ffn_linear_if_twice and self.ffn_last_activation:
            self.activation_last_fn = ActivationLayer(activation=ffn_last_activation)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, f, e = input_shape

        self.ffn_weight_list = [
            self.add_weight(
                name=f'ffn_weight_{i}',
                shape=[f, e],
                initializer=self.ffn_initializer_list[i],
                regularizer=L1L2(0.0, self.ffn_l2),
                trainable=True,
            )
            for i in range(2 if self.ffn_linear_if_twice else 1)
        ]
        if self.ffn_if_bias:
            self.ffn_bias_list = [
                self.add_weight(
                    name=f'ffn_bias_{i}',
                    shape=[f, 1],
                    initializer='zeros',
                    regularizer=L1L2(0.0, self.ffn_l2),
                    trainable=True,
                )
                for i in range(2 if self.ffn_linear_if_twice else 1)
            ]
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = x * self.ffn_weight_list[0] + (self.ffn_bias_list[0] if self.ffn_if_bias else 0.0)
        x = self.bn_fn(x)
        x = self.ln_fn(x)
        x = self.drop_fn(x)
        x = self.activation_fn(x)

        if self.ffn_linear_if_twice:
            x = x * self.ffn_weight_list[1] + (self.ffn_bias_list[1] if self.ffn_if_bias else 0.0)
            if self.ffn_last_activation:
                x = self.activation_last_fn(x)
            if self.ffn_last_factor:
                x = x * self.ffn_last_factor
        return x


class Linear3dParallelLayer(tf.keras.layers.Layer):  # vs Dense or EinsumDense with n-d
    def __init__(self, linear_unit=1, linear_activation=None, linear_dropout=0.0, linear_if_bn=False, linear_if_ln=False,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, linear_if_bias=True, seed=None):
        super().__init__()
        self.linear_unit = linear_unit
        self.linear_if_bias = linear_if_bias
        self.linear_l1 = linear_l1
        self.linear_l2 = linear_l2
        self.linear_initializer = InitializerLayer(
            initializer=linear_initializer,
            activation=linear_activation,
            seed=seed,
        ).get()

        self.activation_fn = ActivationLayer(linear_activation)
        self.bn_fn = BatchNormalization(axis=1) if linear_if_bn else IdentityLayer()
        self.ln_fn = LayerNormalization(axis=1) if linear_if_ln else IdentityLayer()
        self.dropout_fn = Dropout(rate=linear_dropout)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, self.linear_parallel_num, self.inputs_width = input_shape

        self.linear_weight = self.add_weight(
            name='linear_weight',
            shape=[self.linear_parallel_num, self.inputs_width, self.linear_unit],
            initializer=self.linear_initializer,
            regularizer=L1L2(self.linear_l1, self.linear_l2),
            trainable=True,
        )
        self.linear_bias = self.add_weight(
            name='linear_bias',
            shape=[self.linear_parallel_num, self.linear_unit],
            initializer='zeros',
            regularizer=L1L2(self.linear_l1, self.linear_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = tf.einsum('ijk,jkl->ijl', x, self.linear_weight) + (self.linear_bias if self.linear_if_bias else 0.0)
        x = self.bn_fn(x)
        x = self.ln_fn(x)
        x = self.dropout_fn(x)
        x = self.activation_fn(x)  # (batch_size, linear_parallel_num, linear_unit)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'linear_weight': self.linear_weight.numpy(),
                'linear_bias': self.linear_bias.numpy(),
            }
        )
        return config


class DNN3dParallelLayer(tf.keras.layers.Layer):
    def __init__(self, dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_initializer=None, dnn_l2=0.0, dnn_if_bias=True,
                 dnn_if_output2d=False, seed=None):
        super().__init__()
        self.dnn_if_output2d = dnn_if_output2d
        self.dnn_hidden_units = dnn_hidden_units
        self.dnn_length = len(self.dnn_hidden_units)
        self.dnn_dropouts = [dnn_dropout] * self.dnn_length if isinstance(dnn_dropout, float) else dnn_dropout
        self.dnn_bns = [dnn_if_bn] * self.dnn_length if isinstance(dnn_if_bn, bool) else dnn_if_bn
        self.dnn_lns = [dnn_if_ln] * self.dnn_length if isinstance(dnn_if_ln, bool) else dnn_if_ln
        self.dnn_bias = [dnn_if_bias] * self.dnn_length if isinstance(dnn_if_bias, bool) else dnn_if_bias

        self.dnn_fn_list = [
            Linear3dParallelLayer(
                linear_activation=dnn_activation,
                linear_if_bias=self.dnn_bias[i],
                linear_unit=self.dnn_hidden_units[i],
                linear_dropout=self.dnn_dropouts[i],
                linear_if_bn=self.dnn_bns[i],
                linear_if_ln=self.dnn_lns[i],
                linear_l2=dnn_l2,
                linear_initializer=dnn_initializer,
                seed=seed,
            ) for i in range(self.dnn_length)
        ]
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        for dnn_fn in self.dnn_fn_list:
            x = dnn_fn(x)

        if self.dnn_if_output2d:
            x = self.flatten_fn(x)
        return x


class Linear2dParallelLayer(tf.keras.layers.Layer):
    def __init__(self, linear_parallel_num=1, linear_activation=None, linear_if_bias=True,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__()
        if linear_parallel_num < 1:
            raise MLGBError

        self.linear_parallel_num = linear_parallel_num
        self.linear_unit = 1

        self.linear3d_parallel_fn = Linear3dParallelLayer(
            linear_unit=self.linear_unit,
            linear_activation=linear_activation,
            linear_if_bias=linear_if_bias,
            linear_dropout=0.0,
            linear_if_bn=False,
            linear_if_ln=False,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = tf.stack([x] * self.linear_parallel_num, axis=1)  # (batch_size, linear_parallel_num, inputs_width)
        x = self.linear3d_parallel_fn(x)  # (batch_size, linear_parallel_num, linear_unit)
        x = self.flatten_fn(x)
        return x


class ConvolutionLayer(tf.keras.layers.Layer):
    def __init__(self, cnn_conv_mode='Conv1D', cnn_filter_num=64, cnn_kernel_height=32, cnn_kernel_width=1,
                 cnn_activation='tanh', cnn_l2=0.0, cnn_initializer=None,
                 cnn_if_max_pool=True, cnn_pool_size=2, seed=None):
        super().__init__()
        if cnn_conv_mode not in ('Conv1D', 'Conv2D'):
            raise MLGBError

        self.cnn_conv_mode = cnn_conv_mode
        self.cnn_if_max_pool = cnn_if_max_pool

        self.cnn_initializer = InitializerLayer(
            initializer=cnn_initializer,
            activation=cnn_activation,
            seed=seed,
        ).get()
        if self.cnn_conv_mode == 'Conv1D':
            self.conv_fn = Conv1D(
                filters=cnn_filter_num,
                kernel_size=cnn_kernel_height,
                strides=1,
                use_bias=True,
                padding='same',
                data_format='channels_last',
                activation=cnn_activation,
                kernel_initializer=self.cnn_initializer,
                kernel_regularizer=L1L2(0.0, cnn_l2),
            )  # (b, f, e) -> (b, f, cnn_filter_num)
            if self.cnn_if_max_pool:
                self.max_pool_fn = MaxPool1D(
                    pool_size=cnn_pool_size,
                    strides=cnn_pool_size,
                    padding='valid',
                    data_format='channels_last',
                )  # (b, f, cnn_filter_num) -> (b, f // cnn_pool_size, cnn_filter_num)
        else:
            self.conv_fn = Conv2D(
                filters=cnn_filter_num,
                kernel_size=(cnn_kernel_height, cnn_kernel_width),  # (k_h, k_w)
                strides=(1, 1),  # (s_h, s_w)
                use_bias=True,
                padding='same',
                data_format='channels_last',
                activation=cnn_activation,
                kernel_initializer=self.cnn_initializer,
                kernel_regularizer=L1L2(0.0, cnn_l2),
            )  # (b, f, e, 1) -> (b, f, e, cnn_filter_num)
            if self.cnn_if_max_pool:
                self.max_pool_fn = MaxPool2D(
                    pool_size=(cnn_pool_size, 1),  # (k_h, k_w)
                    strides=(cnn_pool_size, 1),  # (s_h, s_w)
                    padding='valid',
                    data_format='channels_last',
                )  # (b, f, e, cnn_filter_num) -> (b, f // cnn_pool_size, e, cnn_filter_num)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank not in (3, 4):
            raise MLGBError
        if self.cnn_conv_mode == 'Conv1D' and input_shape.rank != 3:
            raise MLGBError
        if self.cnn_conv_mode == 'Conv2D' and input_shape.rank != 4:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = self.conv_fn(x)
        x = self.max_pool_fn(x) if self.cnn_if_max_pool else x
        return x


class ConvolutionalNeuralNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, cnn_conv_mode='Conv1D', cnn_filter_nums=(64, 32), cnn_kernel_heights=(64, 32), cnn_kernel_widths=1,
                 cnn_activation='tanh', cnn_l2=0.0, cnn_initializer=None,
                 cnn_if_max_pool=True, cnn_pool_sizes=2, seed=None):
        super().__init__()
        if cnn_conv_mode not in ('Conv1D', 'Conv2D'):
            raise MLGBError

        self.cnn_conv_mode = cnn_conv_mode
        self.cnn_length = len(cnn_filter_nums)
        self.cnn_kernel_heights = [cnn_kernel_heights] * self.cnn_length if isinstance(cnn_kernel_heights, int) else cnn_kernel_heights
        self.cnn_kernel_widths = [cnn_kernel_widths] * self.cnn_length if isinstance(cnn_kernel_widths, int) else cnn_kernel_widths
        self.cnn_pool_sizes = [cnn_pool_sizes] * self.cnn_length if isinstance(cnn_pool_sizes, int) else cnn_pool_sizes

        self.cnn_fn_list = [
            ConvolutionLayer(
                cnn_conv_mode=cnn_conv_mode,
                cnn_filter_num=cnn_filter_nums[i],
                cnn_kernel_height=self.cnn_kernel_heights[i],
                cnn_kernel_width=self.cnn_kernel_widths[i],
                cnn_activation=cnn_activation,
                cnn_l2=cnn_l2,
                cnn_initializer=cnn_initializer,
                cnn_if_max_pool=cnn_if_max_pool,
                cnn_pool_size=self.cnn_pool_sizes[i],
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.cnn_length)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank not in (3, 4):
            raise MLGBError
        if self.cnn_conv_mode == 'Conv1D' and input_shape.rank != 3:
            raise MLGBError
        if self.cnn_conv_mode == 'Conv2D' and input_shape.rank != 4:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        for cnn_fn in self.cnn_fn_list:
            x = cnn_fn(x)
        return x


class GatedRecurrentUnitLayer(tf.keras.layers.Layer):
    def __init__(self, gru_hidden_units=(64, 32),
                 gru_activation='tanh', gru_dropout=0.0, gru_l2=0.0, gru_initializer=None,
                 gru_rct_activation='sigmoid', gru_rct_dropout=0.0, gru_rct_l2=0.0, gru_rct_initializer='orthogonal',
                 gru_if_bias=True, gru_reset_after=True, gru_unroll=False, seed=None):
        super().__init__()
        self.gru_length = len(gru_hidden_units)
        self.gru_initializer_list = [
            InitializerLayer(
                initializer=gru_initializer,
                activation=gru_activation,
                seed=seed + i if isinstance(seed, int) else seed,
            ).get()
            for i in range(self.gru_length)
        ]
        self.gru_rct_initializer_list = [
            InitializerLayer(
                initializer=gru_rct_initializer,
                activation=gru_rct_activation,
                seed=seed + i if isinstance(seed, int) else seed,
            ).get()
            for i in range(self.gru_length)
        ]
        self.gru_fn_list = [
            GRU(
                units=gru_hidden_units[i],
                activation=gru_activation,
                dropout=gru_dropout,
                kernel_initializer=self.gru_initializer_list[i],
                kernel_regularizer=L1L2(0.0, gru_l2),
                recurrent_activation=gru_rct_activation,
                recurrent_dropout=gru_rct_dropout,
                recurrent_initializer=self.gru_rct_initializer_list[i],
                recurrent_regularizer=L1L2(0.0, gru_rct_l2),
                use_bias=gru_if_bias,
                reset_after=gru_reset_after,  # True if cuda or GRU_v1 else False(GRU_v3)
                unroll=gru_unroll,  # False if cuda or small memory else True
                return_sequences=True if i < self.gru_length - 1 else False,  # x[:, -1, :]
            )
            for i in range(self.gru_length)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        for gru_fn in self.gru_fn_list:
            x = gru_fn(x)
        return x


class BiGatedRecurrentUnitLayer(tf.keras.layers.Layer):
    def __init__(self, gru_bi_mode='Frontward', gru_hidden_units=(64, 32),
                 gru_activation='tanh', gru_dropout=0.0, gru_l2=0.0, gru_initializer=None,
                 gru_rct_activation='sigmoid', gru_rct_dropout=0.0, gru_rct_l2=0.0, gru_rct_initializer='orthogonal',
                 gru_if_bias=True, gru_reset_after=True, gru_unroll=False, seed=None):
        super().__init__()
        if gru_bi_mode not in BiGRUModeList:
            raise MLGBError

        self.gru_bi_mode = gru_bi_mode

        self.gru_fn_list = [
            GatedRecurrentUnitLayer(
                gru_hidden_units=gru_hidden_units,
                gru_activation=gru_activation,
                gru_dropout=gru_dropout,
                gru_l2=gru_l2,
                gru_initializer=gru_initializer,
                gru_rct_activation=gru_rct_activation,
                gru_rct_dropout=gru_rct_dropout,
                gru_rct_l2=gru_rct_l2,
                gru_rct_initializer=gru_rct_initializer,
                gru_if_bias=gru_if_bias,
                gru_reset_after=gru_reset_after,
                gru_unroll=gru_unroll,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(1 if self.gru_bi_mode in ('Frontward', 'Backward') else 2)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        if self.gru_bi_mode == 'Frontward':
            x = self.gru_fn_list[0](x)
        elif self.gru_bi_mode == 'Backward':
            x = self.gru_fn_list[0](x[:, ::-1, :])
        elif self.gru_bi_mode == 'Frontward+Backward':
            x = self.gru_fn_list[0](x) + self.gru_fn_list[1](x[:, ::-1, :])
        elif self.gru_bi_mode == 'Frontward-Backward':
            x = self.gru_fn_list[0](x) - self.gru_fn_list[1](x[:, ::-1, :])
        elif self.gru_bi_mode == 'Frontward*Backward':
            x = self.gru_fn_list[0](x) * self.gru_fn_list[1](x[:, ::-1, :])
        elif self.gru_bi_mode == 'Frontward,Backward':
            x = tf.stack([self.gru_fn_list[0](x), self.gru_fn_list[1](x[:, ::-1, :])], axis=1)
        else:
            raise MLGBError

        x = x if self.gru_bi_mode == 'Frontward,Backward' else tf.expand_dims(x, axis=1)
        return x


class CapsuleNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, capsule_num=3, capsule_activation='squash', capsule_l2=0.0, capsule_initializer=None,
                 capsule_interest_num_if_dynamic=False, capsule_input_sequence_pad_mode='pre',
                 capsule_routing_initializer='random_normal', seed=None):
        super().__init__()
        if capsule_input_sequence_pad_mode not in ('pre', 'post'):
            raise MLGBError

        self.capsule_num = capsule_num
        self.capsule_l2 = capsule_l2
        self.capsule_interest_num_if_dynamic = capsule_interest_num_if_dynamic
        self.capsule_input_sequence_pad_mode = capsule_input_sequence_pad_mode
        self.capsule_initializer = InitializerLayer(
            initializer=capsule_initializer,
            activation=capsule_activation,
            seed=seed,
        ).get()
        self.capsule_routing_initializer = InitializerLayer(
            initializer=capsule_routing_initializer,
            activation=None,
            seed=seed,
        ).get()
        self.activation_fn = ActivationLayer(activation=capsule_activation)
        self.mask_fn = MaskLayer(att_if_mask=True)

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, seq_len, embed_dim = input_shape
        if self.capsule_interest_num_if_dynamic:
            seq_len = self.get_dynamic_interest_num(seq_len, embed_dim)
            self.seq_len = seq_len

        self.capsule_bilinear_weight = self.add_weight(
            name='capsule_bilinear_weight',
            shape=[seq_len, embed_dim, embed_dim],
            initializer=self.capsule_initializer,
            regularizer=L1L2(0.0, self.capsule_l2),
            trainable=True,
        )
        self.capsule_routing_weight = self.add_weight(
            name='capsule_routing_weight',
            shape=[1, seq_len, embed_dim],
            initializer=self.capsule_routing_initializer,
            regularizer=L1L2(0.0, self.capsule_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        if self.capsule_interest_num_if_dynamic:
            x = x[:, -self.seq_len:, :] if self.capsule_input_sequence_pad_mode == 'pre' else x[:, :self.seq_len, :]

        w = self.capsule_routing_weight
        w_b = self.capsule_bilinear_weight
        for i in range(self.capsule_num):
            w = self.mask_fn(w)
            w = tf.nn.softmax(w, axis=1)
            x_h = tf.einsum('bfe,fee->bfe', x, w_b) * w  # high_level_capsule: x_h = w * w_b * x
            x_h = self.activation_fn(x_h)  # squash

            w_i = tf.einsum('bfe,fee->bfe', x_h, w_b) * x  # routing_logit: w_i = x_h * w_b * x
            w_i = tf.reduce_sum(w_i, axis=0, keepdims=True)
            w = w + w_i  # if `w` isn't be updated(not bp), it's like `w` of RNN.
            x = x_h
        return x

    def get_dynamic_interest_num(self, seq_len, embed_dim):
        seq_k = max(1, min(seq_len, int(numpy.log2(embed_dim))))
        return seq_k










# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from mlgb.tf.configs import (
    numpy,
    tf,
    Flatten,
)
from mlgb.tf.functions import (
    IdentityLayer,
    MaskLayer,
    FlattenAxesLayer,
)
from mlgb.tf.components.linears import (
    DeepNeuralNetworkLayer,
    DNN3dParallelLayer,
    FeedForwardNetworkLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'SharedBottomLayer',
    'EntireSpaceMultitaskModelLayer',
    'MultigateMixtureOfExpertLayer',
    'ProgressiveLayeredExtractionLayer',
    'ParameterEmbeddingPersonalizedNetworkLayer',
]


class SharedBottomHardLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list,
                 hard_bottom_dnn_hidden_units=(64, 32), hard_tower_dnn_hidden_units=(32,),
                 dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.task_fn_list = task_fn_list
        self.task_num = len(task_fn_list)
        hard_tower_dnn_hidden_units = hard_tower_dnn_hidden_units + (1,)

        self.dnn_bottom_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=hard_bottom_dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.dnn_tower_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=hard_tower_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i + 1 if isinstance(seed, int) else seed,
            )
            for i in range(self.task_num)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = self.dnn_bottom_fn(x)

        outputs = []
        for i in range(self.task_num):
            x_tower = self.dnn_tower_fn_list[i](x)
            x_tower = self.task_fn_list[i](x_tower)
            outputs.append(x_tower)
        return outputs


class SharedBottomSoftLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list,
                 soft_dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=True, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.task_fn_list = task_fn_list
        self.task_num = len(task_fn_list)
        soft_dnn_hidden_units = soft_dnn_hidden_units + (1,)

        self.dnn_parallel_fn = DNN3dParallelLayer(
            dnn_hidden_units=soft_dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            dnn_if_output2d=False,
            seed=seed,
        )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = tf.stack([x] * self.task_num, axis=1)  # (b, task_num, f)
        x = self.dnn_parallel_fn(x)  # (b, task_num, dnn_unit)

        outputs = tf.unstack(x, num=self.task_num, axis=1)
        outputs = [self.task_fn_list[i](outputs[i]) for i in range(self.task_num)]
        return outputs


class SharedBottomLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list, share_mode='SB:hard',
                 hard_bottom_dnn_hidden_units=(64, 32), hard_tower_dnn_hidden_units=(32,), soft_dnn_hidden_units=(64, 32),
                 dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if share_mode not in ('SB:hard', 'SB:soft'):
            raise MLGBError

        if share_mode == 'SB:hard':
            self.sb_fn = SharedBottomHardLayer(
                task_fn_list=task_fn_list,
                hard_bottom_dnn_hidden_units=hard_bottom_dnn_hidden_units,
                hard_tower_dnn_hidden_units=hard_tower_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed,
            )
        else:
            self.sb_fn = SharedBottomSoftLayer(
                task_fn_list=task_fn_list,
                soft_dnn_hidden_units=soft_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed,
            )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        outputs = self.sb_fn(x)
        return outputs


class EntireSpaceMultitaskModelLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list, dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.task_fn_list = task_fn_list
        self.task_num = len(task_fn_list)
        self.dnn_hidden_units = dnn_hidden_units + (1,)

        self.dnn_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=self.dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(self.task_num)
        ]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        outputs = []
        for i in range(self.task_num):
            x_i = self.dnn_fn_list[i](x)
            x_i = self.task_fn_list[i](x_i)
            outputs.append(x_i)

        outputs = list(numpy.cumprod(outputs))  # [ctr, ctr * cvr]
        return outputs


class MultigateMixtureOfExpertLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list, expert_num=8,
                 expert_dnn_hidden_units=(64, 32), gate_dnn_hidden_units=(32,), tower_dnn_hidden_units=(32,),
                 dnn_activation='selu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.expert_num = expert_num
        self.task_fn_list = task_fn_list
        self.task_num = len(task_fn_list)
        self.gate_dnn_hidden_units = gate_dnn_hidden_units
        self.tower_dnn_hidden_units = tower_dnn_hidden_units + (1,)

        self.dnn_experts_fn = DNN3dParallelLayer(
            dnn_hidden_units=expert_dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            dnn_if_output2d=False,
            seed=seed,
        )
        if self.gate_dnn_hidden_units:
            self.dnn_gate_in_fn_list = [
                DeepNeuralNetworkLayer(
                    dnn_hidden_units=gate_dnn_hidden_units,
                    dnn_activation=dnn_activation,
                    dnn_dropout=dnn_dropout,
                    dnn_if_bn=dnn_if_bn,
                    dnn_if_ln=dnn_if_ln,
                    dnn_l2=dnn_l2,
                    dnn_initializer=dnn_initializer,
                    seed=seed + i if isinstance(seed, int) else seed,
                ) for i in range(self.task_num)
            ]
        self.dnn_gate_out_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=(expert_num,),
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(self.task_num)
        ]
        self.dnn_tower_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=self.tower_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(self.task_num)
        ]
        self.mask_fn = MaskLayer(att_if_mask=True)
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_experts = tf.stack([x] * self.expert_num, axis=1)
        x_experts = self.dnn_experts_fn(x_experts)  # (b, experts_num, unit)

        outputs = []
        for i in range(self.task_num):
            x_gate = self.dnn_gate_in_fn_list[i](x) if self.gate_dnn_hidden_units else x
            x_gate = self.dnn_gate_out_fn_list[i](x_gate)
            x_gate = tf.expand_dims(x_gate, axis=2)  # (b, experts_num, 1)
            x_gate = self.mask_fn(x_gate)
            x_gate = tf.nn.softmax(x_gate, axis=1)

            x_tower_in = x_experts * x_gate  #
            x_tower_in = self.flatten_fn(x_tower_in)
            x_tower_out = self.dnn_tower_fn_list[i](x_tower_in)
            x_tower_out = self.task_fn_list[i](x_tower_out)
            outputs.append(x_tower_out)
        return outputs


class ProgressiveLayeredExtractionLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list, tower_dnn_hidden_units=(64, 32),
                 expert_men_num=4, shared_men_num=4, men_hidden_units=(64, 32),
                 dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.task_fn_list = task_fn_list
        self.task_num = len(task_fn_list)
        self.expert_men_num = expert_men_num
        self.shared_men_num = shared_men_num
        self.men_length = len(men_hidden_units)
        self.tower_dnn_hidden_units = tower_dnn_hidden_units + (1,)

        self.men_fn_list = [
            [
                DNN3dParallelLayer(
                    dnn_hidden_units=[men_hidden_units[i]],
                    dnn_activation=dnn_activation,
                    dnn_dropout=dnn_dropout,
                    dnn_if_bn=dnn_if_bn,
                    dnn_if_ln=dnn_if_ln,
                    dnn_l2=dnn_l2,
                    dnn_initializer=dnn_initializer,
                    seed=seed + i + j if isinstance(seed, int) else seed,
                )
                for i in range(self.men_length)
            ]
            for j in range(self.task_num + 1)  # e + s
        ]
        self.cgc_ffn_fn_list = [
            [
                FeedForwardNetworkLayer(
                    ffn_linear_if_twice=False,
                    ffn_if_bias=True,
                    ffn_activation=dnn_activation,
                    ffn_if_bn=dnn_if_bn,
                    ffn_if_ln=dnn_if_ln,
                    ffn_dropout=dnn_dropout,
                    ffn_l2=dnn_l2,
                    ffn_initializer=dnn_initializer,
                    seed=seed + i + j if isinstance(seed, int) else seed,
                )
                for i in range(self.men_length)
            ]
            for j in range(self.task_num + 1)  # e + s
        ]
        self.tower_dnn_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=self.tower_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.task_num)
        ]
        self.flatten_fn_list = [Flatten() for _ in range(self.task_num)]

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_s = tf.stack([x] * self.shared_men_num, axis=1)
        x_e_list = [tf.stack([x] * self.expert_men_num, axis=1)] * self.task_num

        for i in range(self.men_length):
            x_s = self.men_fn_list[-1][i](x_s)
            for j in range(self.task_num):
                x_e = self.men_fn_list[j][i](x_e_list[j])
                x_e_list[j] = self.cgc_e_fn(x_e, x_s, i, j)
            x_s = self.cgc_s_fn(x_e_list, i, -1)

        outputs = []
        for i in range(self.task_num):
            x_i = self.flatten_fn_list[i](x_e_list[i])
            x_i = self.tower_dnn_fn_list[i](x_i)
            x_i = self.task_fn_list[i](x_i)
            outputs.append(x_i)
        return outputs

    def cgc_e_fn(self, x_e, x_s, men_i, task_j):  # CustomizedGateControl
        x = tf.concat([x_e, x_s], axis=1)  # (b, e+s, f == men_unit)
        w = self.cgc_ffn_fn_list[task_j][men_i](x)
        w = tf.nn.softmax(w, axis=2)
        x = x * w
        return x

    def cgc_s_fn(self, x_e_list, men_i, task_j):
        x = tf.concat(x_e_list, axis=1)  # (b, e * task_num, f == men_unit), e = e + s
        w = self.cgc_ffn_fn_list[task_j][men_i](x)
        w = tf.nn.softmax(w, axis=2)
        x = x * w
        return x


class ParameterEmbeddingPersonalizedNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, task_fn_list, domain_num=1, tower_dnn_hidden_units=(64, 32),
                 gnu_factor=2.0, gnu_last_activation='sigmoid', gnu_if_concat_general_inputs=True,
                 dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if not domain_num > 0:
            raise MLGBError
        if not gnu_factor > 0:
            raise MLGBError

        self.task_fn_list = task_fn_list
        self.task_num = len(task_fn_list)
        self.tower_dnn_hidden_units = tower_dnn_hidden_units + (domain_num,)
        self.tower_dnn_length = len(self.tower_dnn_hidden_units)
        self.domain_num = domain_num
        self.gnu_factor = gnu_factor
        self.gnu_if_concat_general_inputs = gnu_if_concat_general_inputs

        self.gnu_0_fn = FeedForwardNetworkLayer(
            ffn_linear_if_twice=True,
            ffn_activation=dnn_activation,
            ffn_last_activation=gnu_last_activation,
            ffn_last_factor=gnu_factor,
        )
        self.gnu_n1_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=[self.tower_dnn_hidden_units[i]],
                dnn_activation=dnn_activation,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_dropout=dnn_dropout,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.tower_dnn_length)
        ]
        self.gnu_n2_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=[self.tower_dnn_hidden_units[i]],
                dnn_activation=gnu_last_activation,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_dropout=dnn_dropout,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.tower_dnn_length)
        ]
        self.tower_dnn_fn_list = [
            [
                DeepNeuralNetworkLayer(
                    dnn_hidden_units=[self.tower_dnn_hidden_units[i]],
                    dnn_activation=dnn_activation,
                    dnn_if_bn=dnn_if_bn,
                    dnn_if_ln=dnn_if_ln,
                    dnn_dropout=dnn_dropout,
                    dnn_l2=dnn_l2,
                    dnn_initializer=dnn_initializer,
                    seed=seed + i + j if isinstance(seed, int) else seed,
                )
                for i in range(self.tower_dnn_length)
            ]
            for j in range(self.task_num)
        ]
        if self.gnu_if_concat_general_inputs:
            self.stop_gradient_fn_list = [IdentityLayer(if_stop_gradient=True) for _ in range(2)]
        self.flatten_fn_list = [Flatten() for _ in range(self.task_num)]
        self.flatten_axes_fn = FlattenAxesLayer(axes=[1, 2])

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise MLGBError

        input_0_shape = tf.TensorShape(input_shape[0]) if not isinstance(input_shape[0], tf.TensorShape) else input_shape[0]
        input_1_shape = tf.TensorShape(input_shape[1]) if not isinstance(input_shape[1], tf.TensorShape) else input_shape[1]
        input_2_shape = tf.TensorShape(input_shape[2]) if not isinstance(input_shape[2], tf.TensorShape) else input_shape[2]

        if input_0_shape.rank not in (2, 3):
            raise MLGBError
        if not (input_0_shape.rank == input_1_shape.rank == input_2_shape.rank == 3):
            raise MLGBError
        if not (input_0_shape[2] == input_1_shape[2] == input_2_shape[2]):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_g, x_d, x_uia = inputs  # (general_f, domain_f, user_item_author_f)

        if self.gnu_if_concat_general_inputs:
            x_g_ep = self.stop_gradient_fn_list[0](x_g)  # EPNet
            x_d = tf.concat([x_d, x_g_ep], axis=1)

        x_d = self.gnu_0_fn(x_d)
        x_g = tf.einsum('bie,bje->bije', x_g, x_d)
        x_g = self.flatten_axes_fn(x_g)

        if self.gnu_if_concat_general_inputs:
            x_g_pp = self.stop_gradient_fn_list[1](x_g)  # PPNet
            x_uia = tf.concat([x_uia, x_g_pp], axis=1)

        x_g = self.flatten_fn_list[0](x_g)
        x_uia = self.flatten_fn_list[1](x_uia)

        x_tower_list = [x_g] * self.task_num
        for i in range(self.tower_dnn_length):
            x_gnu = self.gnu_n_fn(x_uia, i)
            for j in range(self.task_num):
                x_tower = self.tower_dnn_fn_list[j][i](x_tower_list[j])
                x_tower_list[j] = x_tower * x_gnu

        outputs = []
        for i in range(self.task_num):
            if self.domain_num == 1:
                x_i = self.task_fn_list[i](x_tower_list[i])
                outputs.append(x_i)
            else:
                domain_outputs = []
                for j in range(self.domain_num):
                    x_i = self.task_fn_list[i][j](x_tower_list[i])
                    domain_outputs.append(x_i)
                outputs.append(domain_outputs)
        return outputs

    def gnu_n_fn(self, x, i):
        x = self.gnu_n1_fn_list[i](x)
        x = self.gnu_n2_fn_list[i](x)
        x = x * self.gnu_factor
        return x





























# coding=utf-8
# author=uliontse

import tensorflow as tf
if tf.__version__ >= '2.16':
    del tf
    import os
    import sys
    os.environ["TF_USE_LEGACY_KERAS"] = '1'
    sys.stderr.write('Using Keras2 backend.\n')  # `pip install tf-keras~=2.16`

import numpy
import pandas
import tensorflow as tf
from sklearn.utils import compute_class_weight
from sklearn.metrics import confusion_matrix, classification_report

import mlgb
from mlgb import get_model
from mlgb.data import get_binary_label_data
from mlgb.utils import check_filepath


if __name__ == '__main__':
    model_name = 'DeepFM'
    lang = 'tf'
    seed = 0

    print(f'mlgb version: {mlgb.__version__}')
    print(f'tf version: {tf.__version__}')
    print(f'model name: {model_name}')

    # path of save_model:
    tmp_dir = '.tmp'
    model_dir = f'{tmp_dir}/{model_name}_{lang}'
    log_dir = f'{model_dir}/log_dir'
    save_model_dir = f'{model_dir}/save_model'
    check_filepath(tmp_dir, model_dir, log_dir, save_model_dir)

    # get_data:
    two_inputs_models = ['PLM', 'GRU4Rec', 'Caser', 'SASRec', 'BERT4Rec', 'BST', 'DIN', 'DIEN', 'DSIN']
    feature_names, (x_train, y_train), (x_test, y_test) = get_binary_label_data(
        n_samples=int(1e4),
        negative_class_weight=0.9,
        test_size=0.15,
        inputs_if_2_groups=True if model_name in two_inputs_models else False,
        seed=seed,
    )
    print(f'features: {[len(names) for names in feature_names]}')

    class_weight = dict(enumerate(compute_class_weight(class_weight='balanced', classes=numpy.unique(y_train), y=y_train)))
    print(f'class_weight: {class_weight}')

    # y_train = numpy.expand_dims(y_train, axis=-1)
    # y_test = numpy.expand_dims(y_test, axis=-1)

    # train and evaluate:
    model = get_model(
        feature_names=feature_names,
        model_name=model_name,
        task='binary',
        aim='ranking',
        lang='tf',
        seed=seed,

    )
    model.compile(
        loss=tf.losses.BinaryCrossentropy(),
        optimizer=tf.optimizers.Nadam(learning_rate=1e-3),
        metrics=[
            # tfa.metrics.F1Score(num_classes=1, threshold=0.5, average='macro'),  # import tensorflow_addons as tfa
            # tf.metrics.F1Score(threshold=0.5, average='macro'),  # tf.__version__ >= '2.13'
            tf.metrics.AUC(),
        ],
    )
    history = model.fit(
        x=x_train,
        y=y_train,
        batch_size=32,
        epochs=10,
        validation_split=0.15,
        class_weight=class_weight,
        callbacks=[
            tf.keras.callbacks.TensorBoard(log_dir=log_dir),
            tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=5, min_delta=1e-3),
        ]
    )

    model.summary()
    print(pandas.DataFrame(history.history))

    test_evaluate = model.evaluate(x=x_test, y=y_test, batch_size=32, return_dict=True)
    print(test_evaluate)

    y_pred_prob = model.predict(x_test)
    y_pred_prob = numpy.squeeze(y_pred_prob)
    y_pred = numpy.where(y_pred_prob >= 0.5, 1, 0)
    print(y_pred.shape)
    print(y_test.shape)

    print(confusion_matrix(y_test, y_pred))
    print(classification_report(y_test, y_pred))

    # ranking model:
    model.save(filepath=save_model_dir, save_format='tf')
    local_model = tf.keras.models.load_model(filepath=save_model_dir)

    y_local_pred = local_model.predict(x_test)
    y_model_pred = model.predict(x_test)
    print('y_local_pred == y_model_pred:', numpy.allclose(y_local_pred, y_model_pred))
    print(numpy.abs(y_local_pred - y_model_pred).max())





# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.torch.configs import torch
from mlgb.torch.functions import (
    LayerNormalizationLayer,
    ActivationLayer,
    InitializerLayer,
    MaskLayer,
    PositionalEncoding,
)
from mlgb.error import MLGBError


__all__ = [
    'SimpleAttentionLayer',
    'LabelAttentionLayer',
    'TransformerLayer',
    'InteractingLayer',
]


class SimpleAttentionLayer(torch.nn.Module):
    def __init__(self, softmax_axis=1, softmax_if_mask=True, softmax_pre_temperature_ratio=1.0, device='cpu'):
        super().__init__()
        if not (1e-2 <= softmax_pre_temperature_ratio <= 1.0):
            raise MLGBError

        self.softmax_axis = softmax_axis
        self.softmax_pre_temperature_ratio = softmax_pre_temperature_ratio
        self.mask_fn = MaskLayer(att_if_mask=softmax_if_mask, device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if len(x) != 2:
                raise MLGBError
            if x[0].ndim not in (2, 3):
                raise MLGBError
            if x[0].shape != x[1].shape:
                raise MLGBError

            self.scale = x[0].shape[-1] ** 0.5  #
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = [(t.to(device=self.device) if t.device.type != self.device.type else t) for t in x]
        x_q, x_k = x
        x_v = x_k

        w = x_q * x_k / self.scale
        w = w / self.softmax_pre_temperature_ratio  # debug power_p of MIND model.
        w = self.mask_fn(w)
        w = torch.softmax(w, dim=self.softmax_axis)
        x_v = x_v * w
        return x_v


class LabelAttentionLayer(SimpleAttentionLayer):
    def __init__(self, softmax_axis=1, softmax_pre_temperature_ratio=1.0, device='cpu'):
        super().__init__(softmax_axis, True, softmax_pre_temperature_ratio, device)


class MultiHeadAttentionLayer(torch.nn.Module):
    def __init__(self, mha_head_num=8, mha_head_dim=32, mha_if_mask=True, mha_initializer=None, device='cpu'):
        super().__init__()
        if not (mha_head_num > 0 and mha_head_dim > 0):
            raise MLGBError

        self.mha_head_num = mha_head_num
        self.mha_head_dim = mha_head_dim
        self.mha_model_dim = mha_head_num * mha_head_dim
        self.initializer_fn = InitializerLayer(
            initializer=mha_initializer,
            activation=None,
        ).get()
        self.mask_fn = MaskLayer(att_if_mask=mha_if_mask, device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if len(x) != 2:
                raise MLGBError
            if not (x[0].ndim == x[1].ndim == 3):
                raise MLGBError
            if x[0].shape != x[1].shape:
                raise MLGBError

            _, self.fields_width, embed_dim = x[1].shape

            self.qkv_input_weight_list = [
                self.initializer_fn(torch.nn.parameter.Parameter(
                    data=torch.empty(
                        size=[embed_dim, self.mha_model_dim],
                        device=self.device,
                    ),
                    requires_grad=True,
                ))
                for _ in range(3)  # 'qkv'
            ]
            self.qkv_output_weight = self.initializer_fn(torch.nn.parameter.Parameter(
                data=torch.empty(
                    size=[self.mha_model_dim, embed_dim],
                    device=self.device,
                ),
                requires_grad=True,
            ))
            self.built = True
        return

    def forward(self, x):
        # NLP inputs only has 1 vocabulary, 1 sequential feature, sequence_max_length == fields_width.
        # NLP's fields_width << CTR's fields_width, NLP's embed_dim >> CTR's embed_dim,
        # so there are different projection.

        self.build(x)
        x = [(t.to(device=self.device) if t.device.type != self.device.type else t) for t in x]
        x_q, x_k = x
        x_v = x_k  # (b, f, e)

        x_q @= self.qkv_input_weight_list[0]  # (batch_size, fields_width, model_dim)
        x_k @= self.qkv_input_weight_list[1]
        x_v @= self.qkv_input_weight_list[2]

        qkv_4d_shape = [-1, self.fields_width, self.mha_head_num, self.mha_head_dim]
        x_q = torch.transpose(torch.reshape(x_q, shape=qkv_4d_shape), *[1, 2])  # (b, head_num, f, head_dim)
        x_k = torch.transpose(torch.reshape(x_k, shape=qkv_4d_shape), *[1, 2])
        x_v = torch.transpose(torch.reshape(x_v, shape=qkv_4d_shape), *[1, 2])

        x_att = torch.einsum('bhid,bhjd->bhij', x_q, x_k) / (self.mha_head_dim ** 0.5)  # (b, head_num, f, f)
        x_att = self.mask_fn(x_att)
        x_att = torch.softmax(x_att, dim=-1)  #
        x_v = torch.einsum('bhfi,bhfj->bhij', x_att, x_v)  # (b, head_num, f, head_dim)

        qkv_3d_shape = [-1, self.fields_width, self.mha_model_dim]
        x_v = torch.reshape(torch.transpose(x_v, *[1, 2]), shape=qkv_3d_shape)  # (b, f, model_dim)
        x_v @= self.qkv_output_weight  # (b, f, e)
        return x_v


class TransformerFeedForwardNetworkLayer(torch.nn.Module):
    def __init__(self, ffn_activation='gelu', ffn_initializer=None, device='cpu'):
        super().__init__()
        self.activation_fn = ActivationLayer(
            activation=ffn_activation,
            device=device,
        )
        self.initializer_fn = InitializerLayer(
            initializer=ffn_initializer,
            activation=None,
        ).get()
        self.initializer_zeros_fn = InitializerLayer(
            initializer='zeros',
            activation=None,
        ).get()
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            _, f, e = x.shape
            self.ffn_weight_list = [
                self.initializer_fn(torch.nn.parameter.Parameter(
                    data=torch.empty(
                        size=[f, e],
                        device=self.device,
                    ),
                    requires_grad=True,
                ))
                for _ in range(2)
            ]
            self.ffn_bias_list = [
                self.initializer_zeros_fn(torch.nn.parameter.Parameter(
                    data=torch.empty(
                        size=[f, 1],
                        device=self.device,
                    ),
                    requires_grad=True,
                ))
                for _ in range(2)
            ]
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = x * self.ffn_weight_list[0] + self.ffn_bias_list[0]
        x = self.activation_fn(x)
        x = x * self.ffn_weight_list[1] + self.ffn_bias_list[1]
        return x


class TransformerResidualLayer(torch.nn.Module):
    def __init__(self, residual_ln_axis=1, residual_dropout=0.0, device='cpu'):
        super().__init__()
        self.ln_fn = LayerNormalizationLayer(axis=residual_ln_axis, device=device)
        self.drop_fn = torch.nn.Dropout(p=residual_dropout)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if len(x) != 2:
                raise MLGBError
            if not (x[0].ndim == x[1].ndim == 3):
                raise MLGBError
            if x[0].shape != x[1].shape:
                raise MLGBError
    
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = [(t.to(device=self.device) if t.device.type != self.device.type else t) for t in x]
        x, x_stack = x

        x = torch.add(x, x_stack)
        x = self.ln_fn(x)
        x = self.drop_fn(x)
        return x


class TransformerLayer(torch.nn.Module):
    def __init__(self, trm_if_pe=False, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_initializer=None, trm_residual_dropout=0.0,
                 device='cpu'):
        super().__init__()
        self.trm_if_ffn = trm_if_ffn

        self.pe_fn = PositionalEncoding(att_if_pe=trm_if_pe)
        self.mha_fn = MultiHeadAttentionLayer(
            mha_head_num=trm_mha_head_num,
            mha_head_dim=trm_mha_head_dim,
            mha_if_mask=trm_mha_if_mask,
            mha_initializer=trm_mha_initializer,
            device=device,
        )
        if self.trm_if_ffn:
            self.ffn_fn = TransformerFeedForwardNetworkLayer(
                ffn_activation=trm_ffn_activation,
                ffn_initializer=trm_ffn_initializer,
                device=device,
            )
        self.residual_fn_list = torch.nn.ModuleList([
            TransformerResidualLayer(
                residual_ln_axis=1,
                residual_dropout=trm_residual_dropout,
                device=device,
            ) for _ in range(2 if self.trm_if_ffn else 1)
        ])
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if len(x) != 2:
                raise MLGBError
            if not (x[0].ndim == x[1].ndim == 3):
                raise MLGBError
            if x[0].shape != x[1].shape:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = [(t.to(device=self.device) if t.device.type != self.device.type else t) for t in x]
        x_q, x_k = x

        x_q = self.pe_fn(x_q)
        x_k = self.pe_fn(x_k)
        x = self.mha_fn([x_q, x_k])
        x_k = self.residual_fn_list[0]([x, x_k])

        if self.trm_if_ffn:
            x = self.ffn_fn(x_k)
            x_k = self.residual_fn_list[1]([x, x_k])
        return x_k


class InteractingLayer(TransformerLayer):
    def __init__(self, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_initializer=None,
                 trm_residual_dropout=0.0, device='cpu'):
        super().__init__(
            trm_if_pe=False,
            trm_mha_head_num=trm_mha_head_num, trm_mha_head_dim=trm_mha_head_dim,
            trm_mha_if_mask=trm_mha_if_mask, trm_mha_initializer=trm_mha_initializer,
            trm_if_ffn=False,
            trm_residual_dropout=trm_residual_dropout, device=device,
        )









# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.torch.configs import (
    torch,
    InputsModeList,
)
from mlgb.torch.functions import (
    FlattenLayer,
    SparseEmbeddingLayer,
    DenseEmbeddingLayer,
    MultiValuedPoolingLayer,
)
from mlgb.error import MLGBError


__all__ = ['InputsLayer']


class FeatureNamesLayer:
    def __init__(self, inputs_if_multivalued=False, inputs_if_sequential=False):
        self.inputs_if_multivalued = inputs_if_multivalued
        self.inputs_if_sequential = inputs_if_sequential

    def check(self, inputs):
        self.inputs_num = len(inputs)

        if self.inputs_num not in (2, 3, 4):
            raise MLGBError
        if self.inputs_num == 2 and (self.inputs_if_multivalued or self.inputs_if_sequential):
            raise MLGBError
        if self.inputs_num == 3 and (self.inputs_if_multivalued and self.inputs_if_sequential):
            raise MLGBError
        return

    def get(self, inputs):
        self.check(inputs)
        dense_feature_names, sparse_feature_names = inputs[0], inputs[1]

        mv_feature_names, seq_feature_names = None, None
        if self.inputs_num in (3, 4):
            if self.inputs_if_multivalued:
                mv_feature_names = inputs[2]
            if self.inputs_if_sequential:
                seq_feature_names = inputs[-1]
        
        return dense_feature_names, sparse_feature_names, mv_feature_names, seq_feature_names


class SequentialInputsLayer(torch.nn.Module):
    def __init__(self, inputs_if_multivalued=False, inputs_if_sequential=False,
                 inputs_if_embed_dense=False, outputs_dense_if_add_sparse=True,
                 embed_sparse_cate_fn=None, embed_sparse_mv_fn=None, embed_sparse_seq_fn=None,
                 embed_dense_fn=None, pool_mv_fn=None, pool_seq_fn=None,
                 device='cpu'):
        super().__init__()
        self.inputs_if_multivalued = inputs_if_multivalued
        self.inputs_if_sequential = inputs_if_sequential
        self.inputs_if_embed_dense = inputs_if_embed_dense
        self.outputs_dense_if_add_sparse = outputs_dense_if_add_sparse
        self.embed_sparse_cate_fn = embed_sparse_cate_fn
        self.embed_sparse_mv_fn = embed_sparse_mv_fn
        self.embed_sparse_seq_fn = embed_sparse_seq_fn
        self.embed_dense_fn = embed_dense_fn
        self.pool_mv_fn = pool_mv_fn
        self.pool_seq_fn = pool_seq_fn
        self.flatten_fn = FlattenLayer(device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            self.inputs_num = len(x)

            if self.inputs_num not in (2, 3, 4):
                raise MLGBError
            if self.inputs_num == 2 and (self.inputs_if_multivalued or self.inputs_if_sequential):
                raise MLGBError
            if self.inputs_num == 3 and (self.inputs_if_multivalued and self.inputs_if_sequential):
                raise MLGBError

            self.built = True
        return

    def forward(self, x):  # torch, tf216:Tuple[numpy.ndarray, ...], tf210: Tuple[Tensor, ...].
        self.build(x)
        dense_2d_tensor = torch.as_tensor(x[0], dtype=torch.float32, device=self.device)
        sparse_2d_tensor = torch.as_tensor(x[1], dtype=torch.int32, device=self.device)

        embed_cate_2d_tensor, embed_cate_3d_tensor = self.embed_sparse_cate_fn(sparse_2d_tensor)

        if self.inputs_if_embed_dense and embed_cate_3d_tensor.ndim != 3:
            raise MLGBError
        
        if self.inputs_if_embed_dense and embed_cate_3d_tensor.ndim == 3:
            embed_dense_3d_tensor = self.embed_dense_fn(dense_2d_tensor)
            embed_cate_3d_tensor = torch.concat([embed_cate_3d_tensor, embed_dense_3d_tensor], dim=1)

        if self.outputs_dense_if_add_sparse:
            dense_2d_tensor = torch.concat([dense_2d_tensor, embed_cate_2d_tensor], dim=1)

        seq_3d_tensor = None
        if self.inputs_num in (3, 4):
            if self.inputs_if_multivalued:
                mv_tensor = torch.as_tensor(x[2], dtype=torch.int32, device=self.device)
                _, embed_mv_4d_tensor = self.embed_sparse_mv_fn(mv_tensor)
                mv_3d_tensor = self.pool_mv_fn(embed_mv_4d_tensor)
                mv_2d_tensor = self.flatten_fn(mv_3d_tensor)

                if embed_cate_3d_tensor.ndim == 2:
                    embed_cate_3d_tensor = torch.concat([embed_cate_3d_tensor, mv_2d_tensor], dim=1)
                else:
                    embed_cate_3d_tensor = torch.concat([embed_cate_3d_tensor, mv_3d_tensor], dim=1)

                if self.outputs_dense_if_add_sparse:
                    dense_2d_tensor = torch.concat([dense_2d_tensor, mv_2d_tensor], dim=1)

            if self.inputs_if_sequential:
                seq_tensor = torch.as_tensor(x[-1], dtype=torch.int32, device=self.device)  #
                _, embed_seq_4d_tensor = self.embed_sparse_seq_fn(seq_tensor)
                seq_3d_tensor = self.pool_seq_fn(embed_seq_4d_tensor)

        return dense_2d_tensor, embed_cate_3d_tensor, seq_3d_tensor


class FeatureInputsLayer(torch.nn.Module):
    def __init__(self, inputs_if_multivalued=False, inputs_if_sequential=False,
                 inputs_if_embed_dense=False, outputs_dense_if_add_sparse=True,
                 embed_sparse_cate_fn=None, embed_sparse_mv_fn=None, embed_sparse_seq_fn=None,
                 embed_dense_fn=None, pool_mv_fn=None, pool_seq_fn=None,
                 device='cpu'):
        super().__init__()
        self.inputs_if_sequential = inputs_if_sequential
        self.outputs_dense_if_add_sparse = outputs_dense_if_add_sparse
        
        self.inputs_seq_fn = SequentialInputsLayer(
            inputs_if_multivalued=inputs_if_multivalued,
            inputs_if_sequential=inputs_if_sequential,
            inputs_if_embed_dense=inputs_if_embed_dense,
            outputs_dense_if_add_sparse=outputs_dense_if_add_sparse,
            embed_sparse_cate_fn=embed_sparse_cate_fn,
            embed_sparse_mv_fn=embed_sparse_mv_fn,
            embed_sparse_seq_fn=embed_sparse_seq_fn,
            embed_dense_fn=embed_dense_fn,
            pool_mv_fn=pool_mv_fn,
            pool_seq_fn=pool_seq_fn,
            device=device,
        )
        self.flatten_fn = FlattenLayer(device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if len(x) not in (2, 3, 4):
                raise MLGBError

            self.built = True
        return
        
    def forward(self, x):
        self.build(x)
        dense_2d_tensor, embed_cate_3d_tensor, seq_3d_tensor = self.inputs_seq_fn(x)
        
        if self.inputs_if_sequential:
            seq_2d_tensor = self.flatten_fn(seq_3d_tensor)

            if embed_cate_3d_tensor.ndim == 2:
                embed_cate_3d_tensor = torch.concat([embed_cate_3d_tensor, seq_2d_tensor], dim=1)
            else:
                embed_cate_3d_tensor = torch.concat([embed_cate_3d_tensor, seq_3d_tensor], dim=1)

            if self.outputs_dense_if_add_sparse:
                dense_2d_tensor = torch.concat([dense_2d_tensor, seq_2d_tensor], dim=1)
        
        return dense_2d_tensor, embed_cate_3d_tensor


class InputsLayer(torch.nn.Module):
    def __init__(self, feature_names, device='cpu',
                 inputs_mode='Inputs:feature',
                 inputs_if_multivalued=False, inputs_if_sequential=False,
                 inputs_if_embed_dense=False, outputs_dense_if_add_sparse=True,
                 embed_dim=32, embed_initializer=None,
                 embed_2d_dim=None, embed_cate_if_output2d=False,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=2, pool_seq_initializer=None,
                 ):
        super().__init__()
        if inputs_mode not in InputsModeList:
            raise MLGBError

        dense_feature_names, sparse_feature_names, mv_feature_names, seq_feature_names = FeatureNamesLayer(
            inputs_if_multivalued=inputs_if_multivalued, 
            inputs_if_sequential=inputs_if_sequential,
        ).get(feature_names)

        self.embed_cate_fn = SparseEmbeddingLayer(
            sparse_feature_names=sparse_feature_names,
            embed_dim=embed_2d_dim if embed_cate_if_output2d else embed_dim,
            embed_if_output2d=embed_cate_if_output2d,
            device=device,
        )
        self.embed_mv_fn = SparseEmbeddingLayer(
            sparse_feature_names=mv_feature_names,
            embed_dim=embed_dim,
            embed_if_output2d=False,
            device=device,
        ) if inputs_if_multivalued else None
        self.embed_seq_fn = SparseEmbeddingLayer(
            sparse_feature_names=seq_feature_names,
            embed_dim=embed_dim,
            embed_if_output2d=False,
            device=device,
        ) if inputs_if_sequential else None
        self.embed_dense_fn = DenseEmbeddingLayer(
            embed_dim=embed_dim,
            embed_initializer=embed_initializer,
            device=device,
        ) if inputs_if_embed_dense else None
        self.pool_mv_fn = MultiValuedPoolingLayer(
            pool_if_output2d=False,
            pool_mode=pool_mv_mode,
            pool_axis=pool_mv_axis,
            pool_initializer=pool_mv_initializer,
            device=device,
        ) if inputs_if_multivalued else None
        self.pool_seq_fn = MultiValuedPoolingLayer(
            pool_if_output2d=False,
            pool_mode=pool_seq_mode,
            pool_axis=pool_seq_axis,
            pool_initializer=pool_seq_initializer,
            device=device,
        ) if inputs_if_sequential else None

        if inputs_mode == 'Inputs:feature':
            self.input_fn = FeatureInputsLayer(
                inputs_if_multivalued=inputs_if_multivalued,
                inputs_if_sequential=inputs_if_sequential,
                inputs_if_embed_dense=inputs_if_embed_dense,
                outputs_dense_if_add_sparse=outputs_dense_if_add_sparse,
                embed_sparse_cate_fn=self.embed_cate_fn,
                embed_sparse_mv_fn=self.embed_mv_fn,
                embed_sparse_seq_fn=self.embed_seq_fn,
                embed_dense_fn=self.embed_dense_fn,
                pool_mv_fn=self.pool_mv_fn,
                pool_seq_fn=self.pool_seq_fn,
                device=device,
            )
        else:
            self.input_fn = SequentialInputsLayer(
                inputs_if_multivalued=inputs_if_multivalued,
                inputs_if_sequential=inputs_if_sequential,
                inputs_if_embed_dense=inputs_if_embed_dense,
                outputs_dense_if_add_sparse=outputs_dense_if_add_sparse,
                embed_sparse_cate_fn=self.embed_cate_fn,
                embed_sparse_mv_fn=self.embed_mv_fn,
                embed_sparse_seq_fn=self.embed_seq_fn,
                embed_dense_fn=self.embed_dense_fn,
                pool_mv_fn=self.pool_mv_fn,
                pool_seq_fn=self.pool_seq_fn,
                device=device,
            )
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if len(x) not in (2, 3, 4):
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)

        outputs = self.input_fn(x)
        return outputs




# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    numpy,
    tf,
    Embedding,
    Activation,
    BatchNormalization,
    L1L2,
    Flatten,
)
from mlgb.error import MLGBError


__all__ = [
    'IdentityLayer',
    'ActivationLayer',
    'InitializerLayer',
    'OneHotLayer',
    'SparseEmbeddingLayer',
    'DenseEmbeddingLayer',
    'PositionalEncoding',
    'BiasEncoding',
    'MaskLayer',
    'SimplePoolingLayer',
    'MultiValuedPoolingLayer',
    'KMaxPoolingLayer',
    'FlattenAxesLayer',
    'TransposeLayer',
]


class IdentityLayer(tf.keras.layers.Layer):
    def __init__(self, if_stop_gradient=False):
        super().__init__()
        self.if_stop_gradient = if_stop_gradient

    @tf.function
    def call(self, inputs):
        x = inputs
        x = tf.stop_gradient(x) if self.if_stop_gradient else x  # tf.constant(x) without tf.function
        return x


class TransposeLayer(tf.keras.layers.Layer):
    def __init__(self, perm=None):
        super().__init__()
        self.perm = perm

    @tf.function
    def call(self, inputs):
        x = inputs
        x = tf.transpose(x, perm=self.perm)
        return x


class FlattenAxesLayer(tf.keras.layers.Layer):
    def __init__(self, axes=(1, 2)):
        super().__init__()
        if len(axes) != 2:
            raise MLGBError
        if 0 in axes:
            raise MLGBError

        self.axes = axes

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 4:
            raise MLGBError

        _, self.p1, self.p2, self.p3 = input_shape
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        if list(self.axes) == [1, 2]:
            x = tf.reshape(x, shape=[-1, self.p1 * self.p2, self.p3])
        elif list(self.axes) == [2, 3]:
            x = tf.reshape(x, shape=[-1, self.p1, self.p2 * self.p3])
        elif list(self.axes) == [1, 3]:
            x = tf.reshape(x, shape=[-1, self.p1 * self.p3, self.p2])
        else:
            raise MLGBError
        return x


class DiceLayer(tf.keras.layers.Layer):
    def __init__(self, dice_if_official_bn=True):
        super().__init__()
        self.epsilon = 1e-8
        self.alpha = self.add_weight(
            name='alpha',
            shape=(),
            initializer='zeros',
            regularizer=None,
            trainable=True,
        )
        self.official_bn_fn = BatchNormalization(
            axis=-1,
            epsilon=self.epsilon,
            center=False,
            scale=False,
        )
        self.bn_fn = self.official_bn_fn if dice_if_official_bn else self.unofficial_bn_fn

    @tf.function
    def call(self, inputs):
        # swish(x) = silu(x) = sigmoid(x) * x
        # prelu(x) = tf.where(x > 0, x, self.alpha * x)
        # dice(x) = tf.where(x > 0, sigmoid(x) * x, (1 - sigmoid(x)) * self.alpha * x)
        x = inputs
        p = self.bn_fn(x)
        p = tf.nn.sigmoid(p)
        x = p * x + (1 - p) * self.alpha * x
        return x

    def unofficial_bn_fn(self, x):
        diff = x - tf.reduce_mean(x, axis=0, keepdims=True)
        var = tf.math.reduce_variance(x, axis=0, keepdims=True)
        x = diff / tf.sqrt(var + self.epsilon)
        return x


class SquashLayer(tf.keras.layers.Layer):
    def __init__(self):
        super().__init__()
        self.alpha = 1.0
        self.beta = 1e-8

    @tf.function
    def call(self, inputs):
        x = inputs
        l1 = self.p_norm(x, p=1, if_root=True)
        l2 = self.p_norm(x, p=2, if_root=True)
        x = x * (l2 / (l2 + self.alpha)) * (1 / (l1 + self.beta))
        return x

    def p_norm(self, x, p=2, if_root=True):  # tf.norm()
        lp = tf.reduce_sum(tf.abs(x) ** p, axis=None, keepdims=False) ** ((1 / p) if if_root else 1)
        return lp


class ActivationLayer(tf.keras.layers.Layer):
    def __init__(self, activation=None):
        super().__init__()
        self.activation = activation
        if self.activation == 'dice':
            self.activation_fn = DiceLayer()
        elif self.activation == 'squash':
            self.activation_fn = SquashLayer()
        elif isinstance(self.activation, tf.keras.layers.Activation):
            self.activation_fn = self.activation
        else:
            self.activation_fn = Activation(self.activation)

    @tf.function
    def call(self, inputs):
        x = inputs
        x = self.activation_fn(x)
        return x


class InitializerLayer:
    def __init__(self, initializer=None, activation=None, seed=None):
        """
        Paper Team: Xavier Glorot, etc.
        Paper Year: 2010
        Paper Name: <Understanding the difficulty of training deep feedforward neural networks>
        Paper Link: https://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf

        Paper Team: Kaiming He, etc.
        Paper Year: 2015
        Paper Name: <Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification>
        Paper Link: https://arxiv.org/pdf/1502.01852.pdf
        """

        self.initializer = initializer
        self.activation = activation
        self.initializer_map = {
            'glorot_normal': tf.keras.initializers.glorot_normal(seed=seed),  # xavier normal
            'glorot_uniform': tf.keras.initializers.glorot_uniform(seed=seed),
            'xavier_normal': tf.keras.initializers.glorot_normal(seed=seed),  # xavier normal
            'xavier_uniform': tf.keras.initializers.glorot_uniform(seed=seed),
            'he_normal': tf.keras.initializers.he_normal(seed=seed),  # kaiming normal
            'he_uniform': tf.keras.initializers.he_uniform(seed=seed),
            'kaiming_normal': tf.keras.initializers.he_normal(seed=seed),  # kaiming normal
            'kaiming_uniform': tf.keras.initializers.he_uniform(seed=seed),
            'lecun_normal': tf.keras.initializers.lecun_normal(seed=seed),
            'lecun_uniform': tf.keras.initializers.lecun_uniform(seed=seed),
            'random_normal': tf.keras.initializers.random_normal(mean=0.0, stddev=1.0, seed=seed),
            'random_uniform': tf.keras.initializers.random_uniform(seed=seed),
            'truncated_normal': tf.keras.initializers.truncated_normal(mean=0.0, stddev=1.0, seed=seed),
            'orthogonal': tf.keras.initializers.orthogonal(seed=seed),
            'zeros': tf.keras.initializers.zeros(),
            'ones': tf.keras.initializers.ones(),
        }
        self.activation_initializer_map = {
            'relu': self.initializer_map['he_uniform'],
            'gelu': self.initializer_map['he_uniform'],
            'silu': self.initializer_map['he_uniform'],
            'prelu': self.initializer_map['he_uniform'],
            'selu': self.initializer_map['he_uniform'],  # lecun_normal
            'tanh': self.initializer_map['glorot_uniform'],
            'sigmoid': self.initializer_map['glorot_uniform'],
            'dice': self.initializer_map['glorot_normal'],
            'squash': self.initializer_map['random_normal'],
        }

    def get(self):
        if isinstance(self.initializer, (tf.initializers.Initializer, tf.keras.initializers.Initializer)):
            initializer_fn = self.initializer
        elif not (self.initializer in self.initializer_map or self.activation in self.activation_initializer_map):
            initializer_fn = self.initializer_map['he_uniform']  #
        elif self.initializer:
            initializer_fn = self.initializer_map[self.initializer]
        else:
            initializer_fn = self.activation_initializer_map[self.activation]
        return initializer_fn


class OneHotLayer(tf.keras.layers.Layer):
    def __init__(self, sparse_feature_names, onehot_dim=None):
        super().__init__()
        self.sparse_feature_names = sparse_feature_names
        self.onehot_dim = onehot_dim

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.fields_width = input_shape[1]
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        sparse_tensor = inputs

        onehot_tensor_list = []
        for i in range(self.fields_width):
            depth = self.onehot_dim if self.onehot_dim else self.sparse_feature_names[i]['feature_nunique']
            sparse_i_tensor = tf.gather(params=sparse_tensor, indices=tf.constant(i, dtype=tf.int32), axis=1)
            onehot_i_tensor = tf.one_hot(indices=sparse_i_tensor, depth=depth)
            onehot_tensor_list.append(onehot_i_tensor)
        onehot_tensor = tf.concat(onehot_tensor_list, axis=1)
        return onehot_tensor


class SparseEmbeddingLayer(tf.keras.layers.Layer):
    def __init__(self, sparse_feature_names, embed_dim=None, embed_l2=0.0, embed_initializer=None,
                 embed_if_output2d=False, seed=None):
        super().__init__()
        self.sparse_feature_names = sparse_feature_names
        self.embed_dim = embed_dim
        self.embed_if_output2d = embed_if_output2d
        self.embed_initializer = InitializerLayer(
            initializer=embed_initializer,
            activation=None,
            seed=seed,
        ).get()

        self.embed_fn_map = {
            f'{i}': Embedding(
                input_dim=feat_dict['embed_feature_nunique'],
                output_dim=feat_dict['embed_dim'] if not self.embed_dim else self.embed_dim,
                mask_zero=feat_dict['mask_zero'],
                input_length=feat_dict['input_length'],
                embeddings_initializer=self.embed_initializer,
                embeddings_regularizer=L1L2(l1=0.0, l2=embed_l2),
            )
            for i, feat_dict in enumerate(self.sparse_feature_names)
        }  # debug: save_model need str key.

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank not in (2, 3):
            raise MLGBError

        self.fields_width = input_shape[1]
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        sparse_tensor = inputs
        embed_tensor_list = [
            self.embed_fn_map[f'{i}'](tf.gather(params=sparse_tensor, indices=tf.constant(i, dtype=tf.int32), axis=1))
            for i in range(self.fields_width)
        ]
        d23_tensor = tf.concat(embed_tensor_list, axis=1)

        if self.embed_if_output2d:
            return d23_tensor, d23_tensor

        if not self.embed_dim:
            raise MLGBError

        d34_tensor = tf.stack(embed_tensor_list, axis=1)
        return d23_tensor, d34_tensor


class DenseEmbeddingLayer(tf.keras.layers.Layer):
    def __init__(self, embed_dim=32, embed_l2=0.0, embed_initializer=None, seed=None):
        super().__init__()
        if not embed_dim:
            raise MLGBError

        self.embed_dim = embed_dim
        self.embed_l2 = embed_l2
        self.embed_initializer = InitializerLayer(
            initializer=embed_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 2:
            raise MLGBError

        self.embed_weight = self.add_weight(
            name='embed_weight',
            shape=[input_shape[1], self.embed_dim],
            initializer=self.embed_initializer,
            regularizer=L1L2(0.0, self.embed_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = tf.einsum('bf,fe->bfe', x, self.embed_weight)
        return x


class PositionalEncoding(tf.keras.layers.Layer):
    def __init__(self, att_if_pe=True):
        super().__init__()
        self.att_if_pe = att_if_pe
        self.wave_length = 1e4  # 1e4 * pi

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 3:
            raise MLGBError

        _, f, e = input_shape

        pos_f = numpy.arange(f).reshape(-1, 1)
        pos_e = numpy.arange(e) // 2 * 2
        pos_x = pos_f / numpy.power(self.wave_length, pos_e / e)  # (f, e)

        pos_x[:, 0::2] = numpy.sin(pos_x[:, 0::2])
        pos_x[:, 1::2] = numpy.cos(pos_x[:, 1::2])

        self.pe = tf.convert_to_tensor(pos_x.reshape(1, f, e), dtype=tf.float32)
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        if self.att_if_pe:
            x += self.pe
        return x


class BiasEncoding(tf.keras.layers.Layer):
    def __init__(self, if_bias=True, bias_l2=0.0, bias_initializer='zeros', seed=None):
        super().__init__()
        self.if_bias = if_bias
        self.bias_l2 = bias_l2
        self.bias_initializer = InitializerLayer(
            initializer=bias_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank < 2:
            raise MLGBError

        if self.if_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=input_shape[1:],
                initializer='zeros',
                regularizer=L1L2(0.0, self.bias_l2),
                trainable=True,
            )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        if self.if_bias:
            x += self.bias
        return x


class MaskLayer(tf.keras.layers.Layer):
    def __init__(self, att_if_mask=True):
        super().__init__()
        self.min_inf = -4294967295  # float32: -2 ** 32 + 1
        self.att_if_mask = att_if_mask

    @tf.function
    def call(self, inputs):
        x = inputs
        if self.att_if_mask:
            x = tf.where(tf.not_equal(x, 0), x, self.min_inf)
        return x


class SimplePoolingLayer(tf.keras.layers.Layer):
    def __init__(self, pool_mode='Pooling:average', pool_axis=-1, pool_axis_if_keep=False):
        super().__init__()
        if pool_mode not in ('Pooling:max', 'Pooling:average', 'Pooling:sum'):
            raise MLGBError

        self.pool_mode = pool_mode
        self.pool_axis = pool_axis
        self.pool_axis_if_keep = pool_axis_if_keep

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank not in (2, 3, 4):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = self.pool_fn(x, self.pool_mode, self.pool_axis, self.pool_axis_if_keep)
        return x

    def pool_fn(self, x, pool_mode, pool_axis, pool_axis_if_keep):
        pool_fn_map = {
            'Pooling:average': lambda x: tf.reduce_mean(x, axis=pool_axis, keepdims=pool_axis_if_keep),
            'Pooling:sum': lambda x: tf.reduce_sum(x, axis=pool_axis, keepdims=pool_axis_if_keep),
            'Pooling:max': lambda x: tf.reduce_max(x, axis=pool_axis, keepdims=pool_axis_if_keep),
        }
        x = pool_fn_map[pool_mode](x)
        return x


class MultiValuedPoolingLayer(tf.keras.layers.Layer):
    def __init__(self, pool_mode='Attention', pool_axis=2, pool_if_output2d=False, pool_l2=0.0, pool_initializer=None, seed=None):
        super().__init__()
        if pool_mode not in ('Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum'):
            raise MLGBError
        if pool_axis not in (1, 2, 3, -1, -2, -3):
            raise MLGBError

        self.pool_mode = pool_mode
        self.pool_axis = pool_axis
        self.pool_if_output2d = pool_if_output2d
        self.pool_l2 = pool_l2
        self.pool_initializer = InitializerLayer(
            initializer=pool_initializer,
            activation=None,
            seed=seed,
        ).get()
        if self.pool_mode.startswith('Pooling:'):
            self.pool_fn = SimplePoolingLayer(
                pool_mode=pool_mode,
                pool_axis=pool_axis,
            )
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        if input_shape.rank != 4:
            raise MLGBError

        _, self.fields_width, self.sequence_length, _ = input_shape

        if self.pool_mode in ('Weighted', 'Attention'):
            self.pool_weight = self.add_weight(
                name='pool_weight',
                shape=[self.fields_width, self.sequence_length],
                initializer=self.pool_initializer,
                regularizer=L1L2(0.0, self.pool_l2),
                trainable=True,
            )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs  # (batch_size, fields_width, sequence_length, embed_dim)

        if self.pool_mode.startswith('Pooling:'):
            x = self.pool_fn(x)
        elif self.pool_mode in ('Weighted', 'Attention'):
            x_w = tf.einsum('ijkl,jk->ijkl', x, self.pool_weight)
            if self.pool_mode == 'Attention':
                x_w = MaskLayer(att_if_mask=True)(x_w)
                x_w = x * tf.nn.softmax(x_w, axis=self.pool_axis)
            x = tf.reduce_sum(x_w, axis=self.pool_axis, keepdims=False)
        else:
            raise MLGBError

        if self.pool_if_output2d:
            x = self.flatten_fn(x)
        return x  # (b, f, e) or (b, f*e)


class KMaxPoolingLayer(tf.keras.layers.Layer):
    def __init__(self, pool_axis=-1):
        super().__init__()
        if pool_axis == 0:
            raise MLGBError

        self.axis = pool_axis

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape) if not isinstance(input_shape, tf.TensorShape) else input_shape
        self.dim = input_shape.rank
        if self.dim not in (3, 4):
            raise MLGBError

        self.axis = list(range(self.dim))[self.axis]
        self.if_transpose = True if (self.axis != self.dim - 1) else False
        self.perm_dim2axis_dict = {
            '3': {
                '1': [0, 2, 1]
            },
            '4': {
                '1': [0, 3, 2, 1],
                '2': [0, 1, 3, 2]
            }
        }  # debug: save_model need str key.
        self.built = True
        return

    @tf.function
    def call(self, inputs, k=1):
        x = inputs

        if self.if_transpose:
            perm = self.perm_dim2axis_dict[str(self.dim)][str(self.axis)]
            x = tf.transpose(x, perm=perm)
            x = tf.math.top_k(x, k=k, sorted=True)[0]  # axis=-1 only.
            x = tf.transpose(x, perm=perm)  # (b, f, e, 1)
        else:
            x = tf.math.top_k(x, k=k, sorted=True)[0]
        return x






